#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import io
from typing import TypedDict
import logging

from nnf import And
from nnf.dimacs import dump
from pysat.formula import CNF, WCNF

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class Direct(Mapping):
    """
    Maps the problem from nnf to pysat.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["ClassicalSAT", "RandomSAT"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "nnf", "version": "0.4.1"},
            {"name": "python-sat", "version": "1.8.dev13"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this mapping has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def map(self, problem: tuple[And, list], config: Config) -> tuple[WCNF, float]:
        """
        Map from the nnf library into the python-sat library.

        :param problem: SAT problem
        :param config: Config with the parameters specified in Config class
        :return: Mapped problem and the time it took to map it
        """
        start = start_time_measurement()
        hard_constraints, soft_constraints = problem

        # Get number of vars. The union is required in case not all vars are present in either tests/constraints.
        nr_vars = len(hard_constraints.vars().union(And(soft_constraints).vars()))

        # Create a var_labels dictionary that will be used when mapping to pysat
        litdic = {f'L{i - 1}': i for i in range(1, nr_vars + 1)}

        # The most convenient way to map between nnf and pysat was to use the native nnf dump function, which exports
        # the problem as a string, which we can then quickly reload from a buffer.

        # Create buffers for dumping:
        hard_buffer = io.StringIO()
        soft_buffer = io.StringIO()

        # Dump constraints and tests to their respective buffers
        dump(hard_constraints, hard_buffer, var_labels=litdic, mode='cnf')
        # tests have to be conjoined, since we will add them as soft constraints.
        dump(And(soft_constraints), soft_buffer, var_labels=litdic, mode='cnf')

        # Load the cnfs from the buffers:
        hard_cnf = CNF(from_string=hard_buffer.getvalue())
        soft_cnf = CNF(from_string=soft_buffer.getvalue())

        # Create wcnf instance.
        total_wcnf = WCNF()

        # Add hard constraints:
        total_wcnf.extend(hard_cnf)

        # Add soft constraints, with weights.
        total_wcnf.extend(soft_cnf, weights=[1] * len(soft_cnf.clauses))

        logging.info(
            f'Generated pysat wcnf with {len(total_wcnf.hard)} constraints and {len(total_wcnf.soft)} tests.'
        )

        return total_wcnf, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "ClassicalSAT":
            from quark.modules.solvers.ClassicalSAT import ClassicalSAT  # pylint: disable=C0415
            return ClassicalSAT()
        elif option == "RandomSAT":
            from quark.modules.solvers.RandomClassicalSAT import RandomSAT  # pylint: disable=C0415
            return RandomSAT()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")

    def reverse_map(self, solution: list) -> tuple[dict, float]:
        """
        Maps the solution returned by the pysat solver into the reference format.

        :param solution: List containing the solution
        :return: Solution mapped accordingly, time it took to map it
        """
        start = start_time_measurement()
        # converts from (3 / -3) -> (L2 : True / L2: False)
        mapped_sol = {f'L{abs(lit) - 1}': (lit > 0) for lit in solution}
        return mapped_sol, end_time_measurement(start)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from time import sleep
from typing import TypedDict
import logging

import numpy as np
from braket.circuits import Circuit
from braket.aws import AwsDevice
from scipy.optimize import minimize

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class QAOA(Solver):
    """
    QAOA with some parts copied/derived from https://github.com/aws/amazon-braket-examples.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = [
            "LocalSimulator", "arn:aws:braket:::device/quantum-simulator/amazon/sv1",
            "arn:aws:braket:::device/quantum-simulator/amazon/tn1",
            "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony",
            "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3"
        ]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "amazon-braket-sdk", "version": "1.88.2"},
            {"name": "scipy", "version": "1.12.0"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """

        if option == "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony":
            from quark.modules.devices.braket.Ionq import Ionq  # pylint: disable=C0415
            return Ionq("ionQ", "arn:aws:braket:us-east-1::device/qpu/ionq/Harmony")
        elif option == "arn:aws:braket:::device/quantum-simulator/amazon/sv1":
            from quark.modules.devices.braket.SV1 import SV1  # pylint: disable=C0415
            return SV1("SV1", "arn:aws:braket:::device/quantum-simulator/amazon/sv1")
        elif option == "arn:aws:braket:::device/quantum-simulator/amazon/tn1":
            from quark.modules.devices.braket.TN1 import TN1  # pylint: disable=C0415
            return TN1("TN1", "arn:aws:braket:::device/quantum-simulator/amazon/tn1")
        elif option == "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3":
            from quark.modules.devices.braket.Rigetti import Rigetti  # pylint: disable=C0415
            return Rigetti("Rigetti Aspen-9", "arn:aws:braket:us-west-1::device/qpu/rigetti/Aspen-M-3")
        elif option == "LocalSimulator":
            from quark.modules.devices.braket.LocalSimulator import LocalSimulator  # pylint: disable=C0415
            return LocalSimulator("LocalSimulator")
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return: Dictionary of parameter settings
        .. code-block:: python

                    return {
                            "shots": {  # number measurements to make on circuit
                                "values": list(range(10, 500, 30)),
                                "description": "How many shots do you need?"
                            },
                            "opt_method": {
                                "values": ["Powell", "Nelder-Mead"],
                                "description": "Which optimization method do you want?"
                            },
                            "depth": {
                                "values": [3],
                                "description": "Which circuit depth for QAOA do you want?"
                            }
                        }

        """
        return {
            "shots": {  # number of measurements to make on circuit
                "values": list(range(10, 500, 30)),
                "description": "How many shots do you need?"
            },
            "opt_method": {
                "values": ["Powell", "Nelder-Mead"],
                "description": "Which optimization method do you want?"
            },
            "depth": {
                "values": [3],
                "description": "Which circuit depth for QAOA do you want?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            shots: int
            opt_method: str
            depth: int

        """
        shots: int
        opt_method: str
        depth: int

    def run(self, mapped_problem: dict, device_wrapper: any, config: Config, **kwargs: dict) -> tuple[any, float, dict]:
        """
        Run QAOA algorithm on Ising.

        :param mapped_problem: Dict containing problem parameters mapped to the Ising model
        :param device_wrapper: Instance of device
        :param config: Solver configuration settings
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        j = mapped_problem['J']
        if np.any(np.iscomplex(j)):
            logging.warning("The problem matrix of the QAOA solver contains imaginary numbers."
                            "This may lead to an error later in the run.")
        else:
            j = np.real(j)

        # Set up the problem
        n_qubits = j.shape[0]

        # User-defined hypers
        depth = config['depth']
        opt_method = config['opt_method']  # SLSQP, COBYLA, Nelder-Mead, BFGS, Powell, ...

        # Initialize reference solution (simple guess)
        bitstring_init = -1 * np.ones([n_qubits])
        energy_init = np.dot(bitstring_init, np.dot(j, bitstring_init))

        # Set tracker to keep track of results
        tracker = {
            'count': 1,  # Elapsed optimization steps
            'optimal_energy': energy_init,  # Global optimal energy
            'opt_energies': [],  # Optimal energy at each step
            'global_energies': [],  # Global optimal energy at each step
            'optimal_bitstring': bitstring_init,  # Global optimal bitstring
            'opt_bitstrings': [],  # Optimal bitstring at each step
            'costs': [],  # Cost (average energy) at each step
            'res': None,  # Quantum result object
            'params': []  # Track parameters
        }

        # Set options for classical optimization
        options = {'disp': True, 'maxiter': 100}
        # options = {'disp': True, 'ftol': 1e-08, 'maxiter': 100, 'maxfev': 50}  # example options

        ##################################################################################
        # Run QAOA optimization on graph
        ##################################################################################

        logging.info(f"Circuit depth hyperparameter:{depth}")
        logging.info(f"Problem size:{n_qubits}")

        # Kick off training
        start = start_time_measurement()
        _, _, tracker = train(
            device=device_wrapper.get_device(),
            options=options,
            p=depth, ising=j,
            n_qubits=n_qubits,
            n_shots=config['shots'],
            opt_method=opt_method,
            tracker=tracker,
            s3_folder=device_wrapper.s3_destination_folder,
            verbose=True
        )
        time_to_solve = end_time_measurement(start)

        # Log optimized results
        logging.info(f"Optimal energy: {tracker['optimal_energy']}")
        logging.info(f"Optimal classical bitstring: {tracker['optimal_bitstring']}")

        # TODO maybe save this plot
        # plt.plot(cycles, optim_classical)
        # plt.xlabel('optimization cycle')
        # plt.ylabel('best classical minimum')
        # plt.show()

        return tracker['optimal_bitstring'], time_to_solve, {}


# QAOA utils (source:
# https://github.com/aws/amazon-braket-examples/blob/main/examples/hybrid_quantum_algorithms/QAOA/utils_qaoa.py)

# Function to implement ZZ gate using CNOT gates
def zz_gate(q1: any, q2: any, gamma: float) -> Circuit:
    """
    Function that returns a circuit implementing exp(-i \\gamma Z_i Z_j) using CNOT gates if ZZ not supported.

    :param q1: Qubit 1 (control)
    :param q2: Qubit 2 (target)
    :param gamma: Gamma parameter (angle)
    :return: ZZ gate
    """
    # Get a circuit
    circ_zz = Circuit()

    # Construct decomposition of ZZ
    circ_zz.cnot(q1, q2).rz(q2, gamma).cnot(q1, q2)

    return circ_zz


# Function to implement evolution with driver Hamiltonian
def driver(beta: float, n_qubits: int) -> Circuit:
    """
    Returns circuit for driver Hamiltonian U(Hb, beta).

    :param beta: Beta parameter (angle)
    :param n_qubits: Number of qubits
    :return: Circuit with rotated qubits
    """
    # Instantiate circuit object
    circ = Circuit()

    # Apply parametrized rotation around x to every qubit
    for qubit in range(n_qubits):
        gate = Circuit().rx(qubit, 2 * beta)
        circ.add(gate)

    return circ


# Helper function for evolution with cost Hamiltonian
def cost_circuit(gamma: float, ising: np.ndarray, device: AwsDevice) -> Circuit:
    """
    Returns circuit for evolution with cost Hamiltonian.

    :param gamma: Gamma parameter (angle)
    :param ising: Ising matrix
    :param device: Device to run the circuit on
    :return: Circuit representing the cost Hamiltonian
    """
    # Instantiate circuit object
    circ = Circuit()

    # Get all non-zero entries (edges) from Ising matrix
    idx = ising.nonzero()
    edges = list(zip(idx[0], idx[1]))

    # Apply ZZ gate for every edge (with corresponding interaction strength)
    for qubit_pair in edges:
        # Get interaction strength from Ising matrix
        int_strength = ising[qubit_pair[0], qubit_pair[1]]
        # For Rigetti we decompose ZZ using CNOT gates
        if device.name in ["Rigetti", "Aspen-9"]:  # TODO make this more flexible
            gate = zz_gate(qubit_pair[0], qubit_pair[1], gamma * int_strength)
        # Classical simulators and IonQ support ZZ gate
        else:
            gate = Circuit().zz(qubit_pair[0], qubit_pair[1], angle=2 * gamma * int_strength)
        circ.add(gate)

    return circ


# Function to build the QAOA circuit with depth p
def circuit(params: np.array, device: AwsDevice, n_qubits: int, ising: np.ndarray) -> Circuit:
    """
    Function to return the full QAOA circuit; depends on device as ZZ implementation depends on gate set of backend.

    :param params: Array containing the beta and gamma parameters
    :param device: Device to run the circuit on
    :param n_qubits: Number of qubits
    :param ising: Ising matrix
    :return: QAOA Circuit
    """

    # Initialize QAOA circuit with first Hadamard layer
    circ = Circuit()
    x_on_all = Circuit().x(range(0, n_qubits))
    circ.add(x_on_all)
    h_on_all = Circuit().h(range(0, n_qubits))
    circ.add(h_on_all)

    # Setup two parameter families
    circuit_length = int(len(params) / 2)
    gammas = params[:circuit_length]
    betas = params[circuit_length:]

    # Add QAOA circuit layer blocks
    for mm in range(circuit_length):
        circ.add(cost_circuit(gammas[mm], ising, device))
        circ.add(driver(betas[mm], n_qubits))

    return circ


# Function that computes cost function for given params
# pylint: disable=R0917
# pylint: disable=R0913
def objective_function(params: np.array, device: AwsDevice, ising: np.ndarray, n_qubits: int, n_shots: int,
                       tracker: dict, s3_folder: tuple[str, str], verbose: bool) -> float:
    """
    Objective function takes a list of variational parameters as input,
    and returns the cost associated with those parameters.

    :param params: Array containing beta and gamma parameters
    :param device: Device to run the circuit on
    :param ising: Ising matrix
    :param n_qubits: Number of qubits
    :param n_shots: Number of measurements to make on the circuit
    :param tracker: Keeps track of the runs on the circuit
    :param s3_folder: AWS S3 bucket
    :param verbose: Controls degree of detail in logs
    :return: Energy expectation value
    """

    if verbose:
        logging.info("==================================" * 2)
        logging.info(f"Calling the quantum circuit. Cycle: {tracker['count']}")

    # Get a quantum circuit instance from the parameters
    qaoa_circuit = circuit(params, device, n_qubits, ising)

    # Classically simulate the circuit
    # Execute the correct device.run call depending on whether the backend is local or cloud based
    if device.name in ["DefaultSimulator", "StateVectorSimulator"]:
        task = device.run(qaoa_circuit, shots=n_shots)
    else:
        task = device.run(qaoa_circuit, s3_folder, shots=n_shots, poll_timeout_seconds=3 * 24 * 60 * 60)

        # Get ID and status of submitted task
        task_id = task.id
        status = task.state()
        logging.info(f"ID of task: {task_id}")
        logging.info(f"Status of task: {status}")

        # Wait for job to complete
        while status != 'COMPLETED':
            status = task.state()
            logging.info(f"Status: {status}")
            sleep(10)

    # Get result for this task
    result = task.result()
    logging.info(result)

    # Convert results (0 and 1) to ising (-1 and 1)
    meas_ising = result.measurements
    meas_ising[meas_ising == 0] = -1

    # Get all energies (for every shot): (n_shots, 1) vector
    all_energies = np.diag(np.dot(meas_ising, np.dot(ising, np.transpose(meas_ising))))

    # Find minimum and corresponding classical string
    energy_min = np.min(all_energies)
    tracker["opt_energies"].append(energy_min)
    optimal_string = meas_ising[np.argmin(all_energies)]
    tracker["opt_bitstrings"].append(optimal_string)
    logging.info(tracker["optimal_energy"])

    # Store optimal (classical) result/bitstring
    if energy_min < tracker["optimal_energy"]:
        tracker.update({"optimal_energy": energy_min, "optimal_bitstring": optimal_string})

    # Store global minimum
    tracker["global_energies"].append(tracker["optimal_energy"])

    # Energy expectation value
    energy_expect = np.sum(all_energies) / n_shots

    if verbose:
        logging.info(f"Minimal energy: {energy_min}")
        logging.info(f"Optimal classical string: {optimal_string}")
        logging.info(f"Energy expectation value (cost): {energy_expect}")

    # Update tracker
    tracker.update({"count": tracker["count"] + 1, "res": result})
    tracker["costs"].append(energy_expect)
    tracker["params"].append(params)

    return energy_expect


# The function to execute the training: run classical minimization.
# pylint: disable=R0917
def train(device: AwsDevice, options: dict, p: int, ising: np.ndarray, n_qubits: int, n_shots: int, opt_method: str,
          tracker: dict, s3_folder: tuple[str, str], verbose: bool = True) -> tuple[float, np.ndarray, dict]:
    """
    Function to run QAOA algorithm for given, fixed circuit depth p.

    :param device: Device to run the circuit on
    :param options: Dict containing parameters of classical part of the QAOA
    :param p: Circuit depth
    :param ising: Ising matrix
    :param n_qubits: Number of qubits
    :param n_shots: Number of measurements to make on the circuit
    :param opt_method: Controls degree of detail in logs
    :param tracker: Keeps track of the runs on the circuit
    :param s3_folder: AWS S3 bucket
    :param verbose: Controls degree of detail in logs
    :return: Results of the training as a tuple of the energy, the angle and the tracker
    """
    logging.info("Starting the training.")
    logging.info("==================================" * 2)
    logging.info(f"OPTIMIZATION for circuit depth p={p}")

    if not verbose:
        logging.info('Param "verbose" set to False. Will not print intermediate steps.')
        logging.info("==================================" * 2)

    # Initialize
    cost_energy = []

    # Randomly initialize variational parameters within appropriate bounds
    gamma_initial = np.random.uniform(0, 2 * np.pi, p).tolist()
    beta_initial = np.random.uniform(0, np.pi, p).tolist()
    params0 = np.array(gamma_initial + beta_initial)

    # Set bounds for search space
    bnds_gamma = [(0, 2 * np.pi) for _ in range(int(len(params0) / 2))]
    bnds_beta = [(0, np.pi) for _ in range(int(len(params0) / 2))]
    bnds = bnds_gamma + bnds_beta

    tracker["params"].append(params0)
    print(f"Qubit count: {n_qubits}")

    # Run classical optimization (example: method='Nelder-Mead')
    try:
        result = minimize(
            objective_function,
            params0,
            args=(device, ising, n_qubits, n_shots, tracker, s3_folder, verbose),
            options=options,
            method=opt_method,
            bounds=bnds,
        )
    except ValueError as e:
        logging.error(f"The following ValueError occurred in module QAOA: {e}")
        logging.error("The benchmarking run terminates with exception.")
        raise Exception("Please refer to the logged error message.") from e

    # Store result of classical optimization
    result_energy = result.fun
    cost_energy.append(result_energy)
    logging.info(f"Final average energy (cost): {result_energy}")
    result_angle = result.x
    logging.info(f"Final angles: {result_angle}")
    logging.info("Training complete.")

    return result_energy, result_angle, tracker



import unittest
from unittest.mock import patch, MagicMock, mock_open
from pathlib import Path
import os
from datetime import datetime

from quark.BenchmarkManager import BenchmarkManager, Instruction


class TestBenchmarkManager(unittest.TestCase):

    def setUp(self):
        """
        Set up resources before each test.
        """
        self.benchmark_manager = BenchmarkManager()
        self.benchmark_manager.store_dir = "/mock/store"
        self.benchmark_manager.application = MagicMock()
        self.benchmark_manager.application.metrics = MagicMock()

    def test_initialization(self):
        # Reset store_dir before testing
        self.benchmark_manager.store_dir = None

        # Assertions
        self.assertIsNone(self.benchmark_manager.store_dir, "Expected store_dir to be None after initialization.")
        self.assertEqual(
            self.benchmark_manager.results,
            [],
            "Expected results to be an empty list after initialization.")

    @patch("os.path.exists", return_value=True)
    @patch("builtins.open", new_callable=mock_open, read_data='[{"key": "value"}]')
    def test_load_interrupted_results(self, mock_open_file, mock_path_exists):
        self.benchmark_manager.interrupted_results_path = "mock_path"
        results = self.benchmark_manager.load_interrupted_results()
        mock_open_file.assert_called_with("mock_path", encoding="utf-8")
        self.assertEqual(results, [{"key": "value"}])

    @patch("os.path.exists", return_value=False)
    def test_load_interrupted_results_no_file(self, mock_path_exists):
        self.benchmark_manager.interrupted_results_path = "mock_path"
        results = self.benchmark_manager.load_interrupted_results()
        self.assertIsNone(results)

    @patch("BenchmarkManager.Path.mkdir")  # Mock Path.mkdir
    @patch("BenchmarkManager.logging.FileHandler")  # Mock FileHandler
    def test_create_store_dir(self, mock_file_handler, mock_path_mkdir):
        # Mock datetime to control the generated timestamp
        dynamic_now = datetime.today()
        expected_date_str = dynamic_now.strftime("%Y-%m-%d-%H-%M-%S")

        # Call the method under test
        self.benchmark_manager._create_store_dir(store_dir="/mock_dir", tag="test_tag")

        # Dynamically build the expected directory path
        expected_dir = f"/mock_dir/benchmark_runs/test_tag-{expected_date_str}"
        expected_log_file = f"{expected_dir}/logging.log"

        # Assertions to check expected outcomes
        self.assertEqual(self.benchmark_manager.store_dir, expected_dir)
        self.assertTrue(self.benchmark_manager.store_dir.startswith("/mock_dir/benchmark_runs/test_tag-"))
        self.assertTrue(self.benchmark_manager.store_dir.endswith(expected_date_str))
        mock_path_mkdir.assert_called_once_with(parents=True, exist_ok=True)
        mock_file_handler.assert_called_once_with(expected_log_file)

    @patch("logging.FileHandler")
    def test_resume_store_dir(self, mock_file_handler):
        # Ensure the directory exists
        test_dir = "/tmp/existing_dir"
        os.makedirs(test_dir, exist_ok=True)

        try:
            self.benchmark_manager._resume_store_dir(store_dir=test_dir)

            # Assertions
            self.assertEqual(self.benchmark_manager.store_dir, test_dir)
            mock_file_handler.assert_called_once_with(f"{test_dir}/logging.log")

        finally:
            # Cleanup the created directory
            if os.path.exists(test_dir):
                os.rmdir(test_dir)

    @patch("logging.FileHandler")
    @patch("logging.getLogger")
    def test_set_logger(self, mock_get_logger, mock_file_handler):
        """
        Test _set_logger method.
        """
        self.benchmark_manager.store_dir = "/mock/store"
        logger_mock = MagicMock()
        mock_get_logger.return_value = logger_mock

        self.benchmark_manager._set_logger()

        mock_file_handler.assert_called_with("/mock/store/logging.log")
        logger_mock.addHandler.assert_called_once()

    @patch("BenchmarkManager.Path.mkdir")
    @patch("os.path.exists", return_value=True)
    @patch("BenchmarkManager.logging.FileHandler")
    def test_resume_store_dir(self, mock_file_handler, mock_path_exists, mock_path_mkdir):
        store_dir = "/mock_dir"
        self.benchmark_manager._resume_store_dir(store_dir)
        self.assertEqual(self.benchmark_manager.store_dir, store_dir)
        mock_file_handler.assert_called_once_with(f"{store_dir}/logging.log")

    @patch("glob.glob", return_value=["/mock_dir/results_1.json", "/mock_dir/results_2.json"])
    @patch("builtins.open", new_callable=mock_open, read_data='[{"result": "test1"}, {"result": "test2"}]')
    def test_collect_all_results(self, mock_open_file, mock_glob):
        results = self.benchmark_manager._collect_all_results()
        self.assertEqual(len(results), 4, "Expected to collect results from multiple files.")
        self.assertEqual(results[0]["result"], "test1", "Expected the first result to match test1.")
        self.assertEqual(results[2]["result"], "test1", "Expected the first result in the second file to match test1.")

    @patch("builtins.open", new_callable=mock_open)
    @patch("json.dump")
    def test_save_as_json(self, mock_json_dump, mock_open_file):
        mock_results = [{"test": "result1"}, {"test": "result2"}]
        self.benchmark_manager._save_as_json(mock_results)
        mock_open_file.assert_called_once_with(f"{self.benchmark_manager.store_dir}/results.json", 'w')
        mock_json_dump.assert_called_once_with(mock_results, mock_open_file(), indent=2)

    @patch("src.BenchmarkManager.Plotter.visualize_results")
    @patch("src.BenchmarkManager.BenchmarkManager._save_as_json")
    @patch("src.BenchmarkManager.ConfigManager")
    def test_summarize_results(self, mock_config_manager, mock_save_json, mock_visualize):
        self.benchmark_manager.summarize_results(["/mock/dir1", "/mock/dir2"])
        mock_save_json.assert_called()
        mock_visualize.assert_called()

    @patch("glob.glob", return_value=["/mock_dir1/results.json", "/mock_dir2/results.json"])
    @patch("builtins.open", new_callable=mock_open, read_data='[{"result": "mock"}]')
    def test_load_results(self, mock_open_file, mock_glob):
        results = self.benchmark_manager.load_results(["/mock_dir1", "/mock_dir2"])
        self.assertEqual(len(results), 4, "Expected to load results from both directories.")
        self.assertEqual(results[0]["result"], "mock", "Expected the first result to match mock.")

    @patch("src.BenchmarkManager.preprocess")
    @patch("src.BenchmarkManager.postprocess")
    def test_traverse_config(self, mock_postprocess, mock_preprocess):
        # Mock the preprocess function to return expected values
        mock_preprocess.return_value = (Instruction.PROCEED, "processed_input", 0.1)

        mock_postprocess.return_value = (Instruction.PROCEED, "postprocessed_output", 0.2)

        mock_benchmark_record_template = MagicMock()
        mock_benchmark_record_template.copy.return_value = MagicMock()
        self.benchmark_manager.benchmark_record_template = mock_benchmark_record_template
        module = {
            "TestModule": {
                "name": "TestModule",
                "config": {"key": "value"},
                "instance": MagicMock(),
                "submodule": {}
            }
        }
        module["TestModule"]["instance"].metrics = MagicMock()

        instruction, output, benchmark_record = self.benchmark_manager.traverse_config(
            module, "input_data", "/mock/path", 1
        )

        # Assertions
        self.assertEqual(instruction, Instruction.PROCEED, "Expected Instruction.PROCEED to be returned.")
        self.assertEqual(output, "postprocessed_output", "Expected processed output to match mock postprocess return.")
        self.assertIsNotNone(benchmark_record, "Expected a BenchmarkRecord instance.")

    # These tests are commented out because:
    # - The `BenchmarkManager` relies on complex dependencies, including filesystem operations (`Path.mkdir`),
    #   logging configurations (`FileHandler`), and application-specific configurations (`ConfigManager`).
    # - Mocking all these dependencies accurately to test the `orchestrate_benchmark` and `run_benchmark` methods
    #   requires significant effort and a well-structured mocking strategy, which is currently incomplete.
    # - We plan to implement these tests in the future
    # @patch("BenchmarkManager.Path.mkdir")
    # @patch("BenchmarkManager.logging.FileHandler")
    # @patch("BenchmarkManager.ConfigManager")
    # @patch("BenchmarkManager.BenchmarkManager._collect_all_results")
    # def test_orchestrate_benchmark(self, mock_collect_results, mock_config_manager, mock_filehandler, mock_mkdir):
    #     # Mock ConfigManager behavior
    #     mock_config_manager.get_config.return_value = {
    #         "application": {"name": "test_application"}
    #     }
    #     mock_config_manager.get_app.return_value = MagicMock()
    #     mock_config_manager.start_create_benchmark_backlog.return_value = []
    #     mock_config_manager.get_reps.return_value = 1

    #     # Mock mkdir to avoid filesystem errors
    #     mock_mkdir.return_value = None

    #     # Mock FileHandler to avoid file creation errors
    #     mock_filehandler.return_value = MagicMock()
    #     mock_filehandler.return_value.level = 10  # Set a valid integer logging level

    #     # Mock _collect_all_results to return an empty list
    #     mock_collect_results.return_value = []

    #     # Create an instance of BenchmarkManager
    #     benchmark_manager = BenchmarkManager()

    #     # Call orchestrate_benchmark
    #     benchmark_manager.orchestrate_benchmark(mock_config_manager, [{"name": "test"}], "/mock/store_dir")

    #     # Assertions
    #     mock_config_manager.get_config.assert_called_once()
    #     mock_config_manager.save.assert_called_once()
    #     mock_mkdir.assert_called()
    #     mock_filehandler.assert_called_once_with(
    #         "/mock/store_dir/benchmark_runs/test_application-<timestamp>/logging.log"
    #     )
    #     mock_collect_results.assert_called_once()

    # @patch("src.BenchmarkManager.Path.mkdir")
    # @patch("builtins.open", new_callable=mock_open)
    # @patch("src.BenchmarkManager.logging.getLogger")
    # def test_run_benchmark(self, mock_get_logger, mock_open_file, mock_mkdir):
    #     # Set up a mocked logger
    #     mock_logger = MagicMock()
    #     mock_get_logger.return_value = mock_logger

    #     # Mock the BenchmarkManager instance and dependencies
    #     benchmark_manager = BenchmarkManager()
    #     benchmark_manager.application = MagicMock()
    #     benchmark_manager.application.metrics = MagicMock()
    #     benchmark_manager.application.metrics.set_module_config = MagicMock()
    #     benchmark_manager.application.metrics.set_preprocessing_time = MagicMock()
    #     benchmark_manager.application.metrics.add_metric = MagicMock()
    #     benchmark_manager.application.metrics.validate = MagicMock()
    #     benchmark_manager.application.save = MagicMock()
    #     benchmark_manager.benchmark_record_template = MagicMock()
    #     benchmark_manager.store_dir = "/mock/store"

    #     # Set up backlog and repetitions
    #     backlog = [{"config": {"name": "test"}, "submodule": None}]
    #     repetitions = 1

    #     # Run the benchmark
    #     benchmark_manager.run_benchmark(backlog, repetitions)

    #     # Assertions
    #     mock_mkdir.assert_called()
    #     mock_open_file.assert_called_with("/mock/store/benchmark_0/application_config.json", 'w')
    #     mock_logger.info.assert_called()  # Ensure logging calls happen
    #     benchmark_manager.application.save.assert_called()  # Ensure save is called



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import numpy as np
from quark.modules.applications.qml.generative_modeling.training.TrainingGenerative import TrainingGenerative, Core, GPU


class Inference(TrainingGenerative):
    """
    This module executes a quantum circuit with parameters of a pretrained model.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("Inference")

        self.target: np.array
        self.n_states_range: list

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module.
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this circuit.

        :return: Configuration settings for the pretrained model
        .. code-block:: python

            return {
                    "pretrained": {
                        "values": [False],
                        "custom_input": True,
                        "postproc": str,
                        "description": "Please provide the parameters of a pretrained model."
                    }
                }
        """
        return {
            "pretrained": {
                "values": [],
                "custom_input": True,
                "postproc": str,
                "description": "Please provide the parameters of a pretrained model."
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            pretrained: str

        """
        pretrained: str

    def get_default_submodule(self, option: str) -> Core:
        """
        Raises ValueError as this module has no submodules.

        :param option: Option name
        :raises ValueError: If called, since this module has no submodules
        """
        raise ValueError("This module has no submodules.")

    def start_training(self, input_data: dict, config: Config, **kwargs: dict) -> dict:
        """
        Method that uses a pretrained model for inference.

        :param input_data: Dictionary with information needed for inference
        :param config: Inference settings
        :param kwargs: Optional additional arguments
        :return: Dictionary including the information of previous modules as well as of this module
        """
        self.n_states_range = range(2 ** input_data['n_qubits'])
        self.target = np.asarray(input_data["histogram_train"])
        execute_circuit = input_data["execute_circuit"]

        parameters = np.load(config["pretrained"])

        pmfs, samples = execute_circuit([parameters.get() if GPU else parameters])
        pmfs = np.asarray(pmfs)
        samples = (
            self.sample_from_pmf(pmf=pmfs[0], n_shots=input_data["n_shots"])
            if samples is None
            else samples[0]
        )

        loss = self.kl_divergence(pmfs.reshape([-1, 1]), self.target)

        input_data["best_parameter"] = parameters.get() if GPU else parameters
        input_data["inference"] = True
        input_data["KL"] = [loss.get() if GPU else loss]
        input_data["best_sample"] = samples.astype(int).get() if GPU else samples.astype(int)

        return input_data



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class Device(Core, ABC):
    """
    The device class abstracts away details of the physical device, such as submitting a task to the quantum system.
    """

    def __init__(self, device_name: str):
        """
        Constructor method.

        :param device_name: Name of the device
        """
        super().__init__(device_name)
        self.device = None
        self.config = None
        self.device_name = self.name

    def get_parameter_options(self) -> dict:
        """
        Returns the parameters to fine-tune the device.

        Should always be in this format:
        .. code-block:: json

           {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How many reads do you want?"
               }
           }

        :return: Available device settings for this device
        """
        return {}

    def set_config(self, config):
        """
        Sets the device configuration.

        :param config: Configuration settings for the device
        """
        self.config = config

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Returns instance of device class (self) and time it takes to call config.

        :param input_data: Input data (not used)
        :param config: Config for the device
        :param kwargs: Optional keyword arguments
        :return: Output and time needed
        """
        start = start_time_measurement()
        self.config = config
        return self, end_time_measurement(start)

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Returns input data and adds device name to the metrics class instance.

        :param input_data: Input data passed by the parent module
        :param config: Solver config
        :param kwargs: Optional keyword arguments
        :return: Output and time needed
        """
        start = start_time_measurement()
        self.metrics.add_metric("device", self.get_device_name())
        return input_data, end_time_measurement(start)

    def get_device(self) -> any:
        """
        Returns device.

        :return: Instance of the device class
        """
        return self.device

    def get_device_name(self) -> str:
        """
        Returns the device name.

        :return: Name of the device
        """
        return self.device_name



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from quark.modules.Core import Core


class Application(Core, ABC):
    """
    The application component defines the workload, comprising a dataset of increasing complexity, a validation, and an
    evaluation function.
    """

    def __init__(self, application_name: str):
        """
        Constructor method.
        """
        super().__init__(application_name)
        self.application_name = self.name
        self.application = None

    def get_application(self) -> any:
        """
        Gets the application.

        :return: self.application
        """
        return self.application

    @abstractmethod
    def save(self, path: str, iter_count: int) -> None:
        """
        Saves the concrete problem.

        :param path: Path of the experiment directory for this run
        :param iter_count: The iteration count
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

import pandas as pd
from tensorboard.backend.event_processing.event_accumulator import EventAccumulator


class DataHandler(ABC):
    """
    Abstract base class for DataHandler. This class defines the
    necessary methods that both supervised and unsupervised QML applciations
    must implement.
    """

    @abstractmethod
    def data_load(self, gen_mod: dict, config: dict) -> tuple[any, float]:
        """
        Helps to ensure that the model can effectively learn the underlying
        patterns and structure of the data, and produce high-quality outputs.

        :param gen_mod: Dictionary with collected information of the previous modules
        :param config: Config specifying the parameters of the data handler
        :return: Mapped problem and the time it took to create the mapping
        """
        pass

    @abstractmethod
    def evaluate(self, solution: any) -> tuple[any, float]:
        """
        Computes the best loss values.

        :param solution: Solution data
        :return: Evaluation data and the time it took to create it
        """
        pass

    @staticmethod
    def tb_to_pd(logdir: str, rep: str) -> None:
        """
        Converts TensorBoard event files in the specified log directory
        into a pandas DataFrame and saves it as a pickle file.

        :param logdir: Path to the log directory containing TensorBoard event files
        :param rep: Repetition counter
        """
        event_acc = EventAccumulator(logdir)
        event_acc.Reload()
        tags = event_acc.Tags()
        data = []
        tag_data = {}
        for tag in tags['scalars']:
            data = event_acc.Scalars(tag)
            tag_values = [d.value for d in data]
            tag_data[tag] = tag_values
        data = pd.DataFrame(tag_data, index=[d.step for d in data])
        data.to_pickle(f"{logdir}/data_{rep}.pkl")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod

from quark.modules.applications.Application import Application


class QML(Application, ABC):
    """
    qml Module for QUARK, is used by all qml applications.
    """

    @abstractmethod
    def generate_problem(self, config: dict) -> any:
        """
        Creates a concrete problem and returns it.

        :param config: Configuration dictionary
        :return: Generated problem
        """
        pass

    def save(self, path: str, iter_count: int) -> None:
        """
        Placeholder method for saving output to a file.

        :param path: Path to save the file
        :param iter_count: Iteration count
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import math
import numpy as np


class MetricsGeneralization:
    """
    A class to compute generalization metrics for generated samples based on train and solution sets.
    """

    def __init__(self, train_set: np.array, train_size: float, solution_set: np.array, n_qubits: int):
        """
        :param train_set: Set of queries in the training set
        :param train_size: The fraction of queries used for training
        :param solution_set: Set of queries in the solution set
        :param n_qubits: The number of qubits.
        """
        self.train_set = train_set
        self.train_size = train_size
        self.solution_set = solution_set
        self.n_states = 2 ** n_qubits
        self.n_shots = 10000

        self.mask_new, self.mask_sol = self.get_masks()

    def get_masks(self) -> tuple[np.array, np.array]:
        """
        Method to determine the masks, on which the generalization metrics are based.

        :return: Masks needed to determine the generalization metrics for a given train and solution set
        """
        mask_new = np.ones(self.n_states, dtype=bool)
        mask_new[self.train_set] = 0

        mask_sol = np.zeros(self.n_states, dtype=bool)
        mask_sol[self.solution_set] = 1
        mask_sol[self.train_set] = 0

        return mask_new, mask_sol

    def get_metrics(self, generated: np.array) -> dict:
        """
        Method that determines all generalization metrics of a given multiset of generated samples.

        :param generated: Generated samples
        :return: Dictionary with generalization metrics
        """
        g_new = np.sum(generated[self.mask_new])
        g_sol = np.sum(generated[self.mask_sol])
        g_sol_unique = generated[self.mask_sol & self.mask_new]
        g_sol_unique = g_sol_unique[g_sol_unique != 0].size
        g_train = generated[self.train_set]

        results = {
            "fidelity": self.fidelity(g_new, g_sol),
            "exploration": self.exploration(g_new),
            "coverage": self.coverage(g_sol_unique),
            "normalized_rate": self.normalized_rate(g_sol),
            "precision": self.precision(g_sol, g_train)

        }
        return results

    def fidelity(self, g_new: float, g_sol: float) -> float:
        """
        Method to determine the fidelity.

        :param g_new: Multi-subset of unseen queries (noisy or valid)
        :param g_sol: Multi-subset of unseen and valid queries
        :return: Fidelity
        """
        return g_sol / g_new

    def coverage(self, g_sol_unique: float) -> float:
        """
        Method to determine the coverage.

        :param g_sol_unique: Subset of unique unseen and valid queries
        :return: Coverage
        """
        return g_sol_unique / (math.ceil(1 - self.train_size) * len(self.solution_set))

    def normalized_rate(self, g_sol: float) -> float:
        """
        Method to determine the normalized_rate.

        :param g_sol: Multi-subset of unseen and valid queries
        :return: Normalized_rate
        """
        return g_sol / ((1 - self.train_size) * self.n_shots)

    def exploration(self, g_new: float) -> float:
        """
        Method to determine the exploration.

        :param g_new: Multi-subset of unseen queries (noisy or valid)
        :return: Exploration
        """
        return g_new / self.n_shots

    def precision(self, g_sol: float, g_train: float) -> float:
        """
        Method to determine the precision.

        :param g_sol: Multi-subset of unseen and valid queries
        :param g_train: Number of queries that were memorized from the training set
        :return: Precision
        """
        return (np.sum(g_sol) + np.sum(g_train)) / self.n_shots



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

import dwave_networkx as dnx
import networkx

from quark.modules.applications.Mapping import Mapping, Core
from quark.utils import start_time_measurement, end_time_measurement


class QUBO(Mapping):
    """
    QUBO formulation for the TSP.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Annealer"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "networkx", "version": "3.4.2"},
            {"name": "dwave_networkx", "version": "0.8.15"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this mapping.

        :return: Dictionary with configurable settings
        .. code-block:: python

            return {
                    "lagrange_factor": {
                                        "values": [0.75, 1.0, 1.25],
                                        "description": "By which factor would you like to multiply your "
                                                        "Lagrange?",
                                        "custom_input": True,
                                        "postproc": float
                    }
                }
        """
        return {
            "lagrange_factor": {
                "values": [0.75, 1.0, 1.25],
                "description": "By which factor would you like to multiply your Lagrange?",
                "custom_input": True,
                "allow_ranges": True,
                "postproc": float  # Since we allow custom input here we need to parse it to float (input is str)
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             lagrange_factor: float

        """
        lagrange_factor: float

    def map(self, problem: networkx.Graph, config: Config) -> tuple[dict, float]:
        """
        Maps the networkx graph to a QUBO formulation.

        :param problem: Networkx graph
        :param config: Config with the parameters specified in Config class
        :return: Dict with QUBO, time it took to map it
        """
        start = start_time_measurement()
        lagrange = None
        lagrange_factor = config['lagrange_factor']
        weight = 'weight'

        # Taken from dwave_networkx.traveling_salesperson_qubo
        lagrange = problem.size(weight=weight) * problem.number_of_nodes() / problem.number_of_edges()

        lagrange = lagrange * lagrange_factor

        logging.info(f"Default Lagrange parameter: {lagrange}")

        # Get a QUBO representation of the problem
        q = dnx.traveling_salesperson_qubo(problem, lagrange, weight)

        return {"Q": q}, end_time_measurement(start)

    def get_default_submodule(self, option: str) -> Core:
        """
        Get the default submodule based on the given option.

        :param option: Submodule option
        :return: Corresponding submodule
        :raises NotImplemented: If the provided option is not implemented
        """

        if option == "Annealer":
            from quark.modules.solvers.Annealer import Annealer  # pylint: disable=C0415
            return Annealer()
        else:
            raise NotImplementedError(f"Solver Option {option} not implemented")



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

import nnf
import numpy as np
from nnf import Var, And, Or
from nnf.dimacs import dump

from quark.modules.Core import Core
from quark.modules.applications.optimization.Optimization import Optimization
from quark.utils import start_time_measurement, end_time_measurement


class SAT(Optimization):
    """
    The SAT (Satisfiability) problem plays a crucial role in the field of computational optimization. In the context
    of vehicle manufacturing, it is essential to test various pre-series vehicle configurations to ensure they meet
    specific requirements before production begins. This testing involves making sure that each vehicle configuration
    complies with several hard constraints related to safety, performance, and buildability while also fulfilling
    soft constraints such as feature combinations or specific requirements for testing. The SAT problem models these
    constraints in a way that enables a systematic approach to determine feasible vehicle configurations and minimize
    the need for excessive physical prototypes.

    This problem is modeled as a Max-SAT problem, where the aim is to find a configuration that satisfies as many
    constraints as possible while balancing between the number of satisfied hard and soft constraints. The formulation
    uses a conjunctive normal form (CNF) representation of logical expressions to model the dependencies and
    incompatibilities between various features and components in vehicle assembly. By leveraging optimization
    algorithms, the SAT module aims to produce a minimal but sufficient set of configurations, ensuring that all
    necessary tests are performed while minimizing resource usage. This approach helps in creating a robust testing
    framework and reducing the overall cost of vehicle development.

    To solve the SAT problem, various approaches are employed, including translating the CNF representation into
    different quantum and classical optimization mappings such as QUBO (Quadratic Unconstrained Binary Optimization)
    or Ising formulations. These mappings make the SAT problem suitable for solving on quantum computers and
    classical annealers. The SAT problem in this module is implemented with a flexible interface, allowing integration
    with a range of solvers that can exploit different computational paradigms, making it adaptable for a variety of
    hardware and optimization backends.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("SAT")
        self.submodule_options = [
            "QubovertQUBO", "Direct", "ChoiQUBO", "DinneenQUBO", "ChoiIsing", "DinneenIsing"
        ]
        self.literals = None
        self.num_tests = None
        self.num_constraints = None
        self.num_variables = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "nnf", "version": "0.4.1"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_solution_quality_unit(self) -> str:
        return "Evaluation"

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: Option specifying the submodule
        :return: Instance of the corresponding submodule
        :raises NotImplementedError: If the option is not recognized
        """
        if option == "QubovertQUBO":
            from quark.modules.applications.optimization.SAT.mappings.QubovertQUBO import \
                QubovertQUBO  # pylint: disable=C0415
            return QubovertQUBO()
        elif option == "Direct":
            from quark.modules.applications.optimization.SAT.mappings.Direct import Direct  # pylint: disable=C0415
            return Direct()
        elif option == "ChoiQUBO":
            from quark.modules.applications.optimization.SAT.mappings.ChoiQUBO import ChoiQUBO  # pylint: disable=C0415
            return ChoiQUBO()
        elif option == "ChoiIsing":
            from quark.modules.applications.optimization.SAT.mappings.ChoiISING import ChoiIsing  # pylint: disable=C0415
            return ChoiIsing()
        elif option == "DinneenQUBO":
            from quark.modules.applications.optimization.SAT.mappings.DinneenQUBO import DinneenQUBO  # pylint: disable=C0415
            return DinneenQUBO()
        elif option == "DinneenIsing":
            from quark.modules.applications.optimization.SAT.mappings.DinneenISING import \
                DinneenIsing  # pylint: disable=C0415
            return DinneenIsing()
        else:
            raise NotImplementedError(f"Mapping Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Dictionary with configurable settings
        .. code-block:: python

            return {
                    "variables": {
                        "values": list(range(10, 151, 10)),
                        "custom_input": True,
                        "allow_ranges": True,
                        "postproc": int,
                        "description": "How many variables do you need?"
                    },
                    "clvar_ratio_cons": {
                        "values": [2, 3, 4, 4.2, 5],
                        "custom_input": True,
                        "allow_ranges": True,
                        "postproc": int,
                        "description": "What clause-to-variable ratio do you want for the (hard) constraints?"
                    },
                    "clvar_ratio_test": {
                        "values": [2, 3, 4, 4.2, 5],
                        "custom_input": True,
                        "allow_ranges": True,
                        "postproc": int,
                        "description": "What clause-to-variable ratio do you want for the tests (soft con.)?"
                    },
                    "problem_set": {
                        "values": list(range(10)),
                        "description": "Which problem set do you want to use?"
                    },
                    "max_tries": {
                        "values": [100],
                        "description": "Maximum number of tries to create problem?"
                    }
                }
        """
        return {
            "variables": {
                "values": list(range(10, 101, 10)),
                "custom_input": True,
                "allow_ranges": True,
                "postproc": int,
                "description": "How many variables do you need?"
            },
            "clvar_ratio_cons": {
                "values": [2, 3, 4, 4.2, 5],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": int,
                "description": "What clause-to-variable ratio do you want for the (hard) constraints?"
            },
            "clvar_ratio_test": {
                "values": [2, 3, 4, 4.2, 5],
                "custom_input": True,
                "allow_ranges": True,
                "postproc": int,
                "description": "What clause-to-variable ratio do you want for the tests (soft constraints)?"
            },
            "problem_set": {
                "values": list(range(10)),
                "description": "Which problem set do you want to use?"
            },
            "max_tries": {
                "values": [100],
                "description": "Maximum number of tries to create problem?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            variables: int
            clvar_ratio_cons: float
            clvar_ratio_test: float
            problem_set: int
            max_tries: int

        """
        variables: int
        clvar_ratio_cons: float
        clvar_ratio_test: float
        problem_set: int
        max_tries: int

    def generate_problem(self, config: Config) -> tuple[nnf.And, list]:
        """
        Generates a vehicle configuration problem out of a given config.
        Returns buildability constraints (hard constraints) and tests (soft
        constraints), the successful evaluation of which we try to maximize.
        Both are given in nnf form, which we then convert accordingly.

        :param config: Configuration parameters for problem generation
        :return: A tuple containing the problem, number of variables, and other details
        """

        self.num_variables = config["variables"]
        num_constraints = round(config["clvar_ratio_cons"] * self.num_variables)
        num_tests = round(config["clvar_ratio_test"] * self.num_variables)
        max_tries = config["max_tries"]
        self.literals = [Var(f"L{i}") for i in range(self.num_variables)]
        self.application = {}

        def _generate_3sat_clauses(nr_clauses, nr_vars, satisfiable, rseed, nr_tries):
            # Iterate over the desired number of attempts: break if we find a solvable instance.
            for attempt in range(nr_tries):
                # Initialize random number generator -- multiply the attempt to traverse distinct random seeds
                # for the hard and soft constraints, respectively (since rseed of the hard and soft constraints differs
                # by 1).
                rng = np.random.default_rng(rseed + attempt * 2)
                clause_list = []
                # generate literal list to sample from
                lit_vars = [Var(f"L{i}") for i in range(nr_vars)]
                for _ in range(nr_clauses):
                    # Select three (non-repeated) literals and negate them randomly -- together constituting a clause
                    chosen_literals = rng.choice(lit_vars, 3, replace=False)
                    negate_literals = rng.choice([True, False], 3, replace=True)
                    # Perform the random negations and append to clause:
                    clause = [
                        lit.negate() if neg else lit
                        for lit, neg in zip(chosen_literals, negate_literals)
                    ]
                    # Append the generated clause to the total container
                    clause_list.append(Or(clause))
                prob = And(clause_list)
                if not satisfiable or prob.satisfiable():
                    return clause_list

            # Loop ran out of tries
            logging.error("Unable to generate valid solutions. Consider increasing max_tries or decreasing "
                          "the clause:variable ratio.")
            raise ValueError("Unable to generate valid solution.")

        # Choose a random seed -- since we try at most max_tries times to generate a solvable instance,
        # Space the initial random seeds by 2 * max_tries (because we need both hard and soft constraints).
        random_seed = 2 * config["problem_set"] * max_tries
        # Generate hard  & soft constraints. Make both satisfiable, but this can in principle be tuned.
        hard = And(_generate_3sat_clauses(
            num_constraints, self.num_variables, satisfiable=True,
            rseed=random_seed, nr_tries=max_tries
        ))
        # The random_seed + 1 ensures that a different set of seeds is sampled compared to the hard constraints.
        soft = _generate_3sat_clauses(
            num_tests, self.num_variables, satisfiable=True,
            rseed=random_seed + 1, nr_tries=config["max_tries"]
        )
        if (hard is None) or (soft is None):
            raise ValueError("Unable to generate satisfiable")
        # Saving constraints and tests
        self.application["constraints"] = hard
        self.application["tests"] = soft
        # And their cardinalities:
        self.num_constraints = len(hard)
        self.num_tests = len(soft)

        logging.info(f"Generated a vehicle options Max3SAT"
                     f" instance with {self.num_variables} variables, {self.num_constraints} constraints"
                     f" and {self.num_tests} tests")
        return hard, soft

    def validate(self, solution: dict) -> tuple[bool, float]:
        """
        Validate a given solution against the constraints.

        :param solution: The solution to validate
        :return: True if the solution is valid, False otherwise, and time it took to complete
        """
        start = start_time_measurement()

        logging.info("Checking validity of solution:")
        nr_satisfied_hardcons = len(*np.where(
            [c.satisfied_by(solution) for c in self.application["constraints"].children]
        ))
        ratio = nr_satisfied_hardcons / self.num_constraints
        is_valid = ratio == 1.0
        logging.info(f"Ratio of satisfied constraints: {ratio}\nSuccess:{['no', 'yes'][int(is_valid)]}")

        return is_valid, end_time_measurement(start)

    def evaluate(self, solution: dict) -> tuple[float, float]:
        """
        Calculates the quality of the solution.

        :param solution: Dictionary containing the solution
        :return: Tour length, time it took to calculate the tour length
        """
        start = start_time_measurement()
        logging.info("Checking the quality of the solution:")

        # Count the number of satisfied clauses
        nr_satisfied_tests = len(*np.where([test.satisfied_by(solution) for test in self.application["tests"]]))

        ratio_satisfied = nr_satisfied_tests / self.num_tests
        logging.info(f"Ratio of satisfied test clauses: {ratio_satisfied}.")

        return ratio_satisfied, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Save the constraints and tests to files in CNF format.

        :param path: The directory path where the files will be saved.
        :param iter_count: The iteration count to include in the filenames.
        """
        with open(f"{path}/constraints_iter_{iter_count}.cnf", "w") as f_cons:
            dump(
                obj=self.application["constraints"],
                fp=f_cons,
                var_labels={str(literal): idx + 1 for idx, literal in enumerate(self.literals)}
            )
        with open(f"{path}/tests_iter_{iter_count}.cnf", "w") as f_test:
            dump(
                obj=Or(self.application["tests"]),
                fp=f_test,
                var_labels={str(literal): idx + 1 for idx, literal in enumerate(self.literals)}
            )



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC
import logging
from typing import TypedDict

from quark.utils import start_time_measurement, end_time_measurement
from quark.modules.Core import Core
from quark.modules.applications.qml.Model import Model


class LibraryGenerative(Core, Model, ABC):
    """
    This class is an abstract base class for mapping a library-agnostic gate sequence to a library such as Qiskit.
    It provides no concrete implementations of abstract methods and is intended to be extended by specific libraries.
    """

    def __init__(self, name: str):
        """
        Constructor method.

        :param name: Name of the model
        """
        self.name = name
        super().__init__()

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             backend: str
             n_shots: int

        """
        backend: str
        n_shots: int

    def preprocess(self, input_data: dict, config: Config, **kwargs) -> tuple[dict, float]:
        """
        Base class for mapping the gate sequence to a library such as Qiskit.

        :param input_data: Collection of information from the previous modules
        :param config: Config specifying the number of qubits of the circuit
        :param kwargs: Optional keyword arguments
        :return: Tuple including dictionary with the function to execute the quantum circuit on a simulator or quantum
                 hardware and the computation time of the function
        """
        start = start_time_measurement()

        output = self.sequence_to_circuit(input_data)
        backend = self.select_backend(config["backend"], output["n_qubits"])
        output["execute_circuit"], output['circuit_transpiled'] = self.get_execute_circuit(
            output["circuit"],
            backend,
            config["backend"],
            config
        )
        output["backend"] = config["backend"]
        output["n_shots"] = config["n_shots"]
        logging.info("Library created")
        output["store_dir"] = kwargs["store_dir"]

        return output, end_time_measurement(start)

    def postprocess(self, input_data: dict, config: dict, **kwargs) -> tuple[dict, float]:
        """
        This method corresponds to the identity and passes the information of the subsequent module
        back to the preceding module in the benchmarking process.

        :param input_data: Collected information of the benchmarking procesS
        :param config: Config specifying the number of qubits of the circuit
        :param kwargs: Optional keyword arguments
        :return: Tuple with input dictionary and the computation time of the function
        """
        start = start_time_measurement()
        return input_data, end_time_measurement(start)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import pickle
import logging
import os

import networkx as nx
import matplotlib.pyplot as plt
import numpy as np

from quark.modules.applications.Application import Core
from quark.modules.applications.optimization.Optimization import Optimization
from quark.utils import start_time_measurement, end_time_measurement


class TSP(Optimization):
    """
    "The famous travelling salesman problem (also called the travelling salesperson problem or in short TSP) is a
    well-known NP-hard problem in combinatorial optimization, asking for the shortest possible route that visits each
    node exactly once, given a list of nodes and the distances between each pair of nodes. Applications of the
    TSP can be found in planning, logistics, and the manufacture of microchips. In these applications, the general
    concept of a node represents, for example, customers, or points on a chip.

    TSP as graph problem: The solution to the TSP can be viewed as a specific ordering of the vertices in a weighted
    graph. Taking an undirected weighted graph, nodes correspond to the graph's nodes, with paths corresponding to the
    graph's edges, and a path's distance is the edge's weight."
    (source: https://github.com/aws/amazon-braket-examples/tree/main/examples)
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__("TSP")
        self.submodule_options = [
            "Ising", "QUBO", "GreedyClassicalTSP", "ReverseGreedyClassicalTSP", "RandomTSP"
        ]

    @staticmethod
    def get_requirements() -> list:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "networkx", "version": "3.4.2"},
            {"name": "numpy", "version": "1.26.4"}
        ]

    def get_solution_quality_unit(self) -> str:
        """
        Returns the unit of measurement for the solution quality.

        :return: Unit of measurement for the solution quality
        """
        return "Tour cost"

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the given option.

        :param option: The chosen submodule option
        :return: The corresponding submodule instance
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "Ising":
            from quark.modules.applications.optimization.TSP.mappings.ISING import Ising  # pylint: disable=C0415
            return Ising()
        elif option == "QUBO":
            from quark.modules.applications.optimization.TSP.mappings.QUBO import QUBO  # pylint: disable=C0415
            return QUBO()
        elif option == "GreedyClassicalTSP":
            from quark.modules.solvers.GreedyClassicalTSP import GreedyClassicalTSP  # pylint: disable=C0415
            return GreedyClassicalTSP()
        elif option == "ReverseGreedyClassicalTSP":
            from quark.modules.solvers.ReverseGreedyClassicalTSP import ReverseGreedyClassicalTSP  # pylint: disable=C0415
            return ReverseGreedyClassicalTSP()
        elif option == "RandomTSP":
            from quark.modules.solvers.RandomClassicalTSP import RandomTSP  # pylint: disable=C0415
            return RandomTSP()
        else:
            raise NotImplementedError(f"Mapping Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application

        :return: Dictionary with configurable settings.
        .. code-block:: python

            return {
                    "nodes": {
                        "values": list([3, 4, 6, 8, 10, 14, 16]),
                        "allow_ranges": True,
                        "description": "How many nodes does your graph need?",
                        "postproc": int
                    }
                }
        """
        return {
            "nodes": {
                "values": list([3, 4, 6, 8, 10, 14, 16]),
                "allow_ranges": True,
                "description": "How many nodes does you graph need?",
                "postproc": int  # postproc needed to parse the result from allow_ranges to int
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

             nodes: int

        """
        nodes: int

    @staticmethod
    def _get_tsp_matrix(graph: nx.Graph) -> np.ndarray:
        """
        Creates distance matrix out of given coordinates.

        :param graph: The input graph
        :return: Distance matrix
        """
        number_of_nodes = len(graph)
        matrix = np.zeros((number_of_nodes, number_of_nodes))
        for i in nx.all_pairs_dijkstra_path_length(graph, weight="weight"):
            distance_dist = i[1]
            for j in distance_dist.items():
                matrix[i[0] - 1][j[0] - 1] = j[1]
                matrix[j[0] - 1][i[0] - 1] = matrix[i[0] - 1][j[0] - 1]

        return matrix

    def generate_problem(self, config: Config) -> nx.Graph:
        """
        Uses the reference graph to generate a problem for a given config.

        :param config: Configuration dictionary
        :return: Graph with the problem
        """

        if config is None:
            config = {"nodes": 5}

        nodes = config['nodes']

        # Read in the original graph
        with open(os.path.join(os.path.dirname(__file__), "data", "reference_graph.gpickle"), "rb") as file:
            graph = pickle.load(file)

        # Remove seams until the target number of seams is reached
        nodes_in_graph = list(graph.nodes)
        nodes_in_graph.sort()

        if len(nodes_in_graph) < nodes:
            raise ValueError("Too many nodes! The original graph has less seams than that!")

        unwanted_nodes = nodes_in_graph[-len(nodes_in_graph) + nodes:]
        unwanted_nodes = [x for x in graph.nodes if x in unwanted_nodes]

        # Remove one node after another
        for node in unwanted_nodes:
            graph.remove_node(node)

        if not nx.is_connected(graph):
            logging.error("Graph is not connected!")
            raise ValueError("Graph is not connected!")

        # Normalize graph
        cost_matrix = self._get_tsp_matrix(graph)
        graph = nx.from_numpy_array(cost_matrix)

        self.application = graph

        return graph

    def process_solution(self, solution: dict) -> tuple[list, float]:
        """
        Convert dict to list of visited nodes.

        :param solution: Dictionary with solution
        :return: Processed solution and the time it took to process it
        """
        start_time = start_time_measurement()
        nodes = self.application.nodes()
        start = np.min(nodes)
        # fill route with None values
        route: list = [None] * len(self.application)

        # Get nodes from sample
        logging.info(str(solution.items()))

        for (node, timestep), val in solution.items():
            if val:
                logging.info((node, timestep))
            if val and (node not in route):
                route[timestep] = node

        # Check whether every timestep has only 1 node flagged
        for i in nodes:
            relevant_nodes = []
            relevant_timesteps = []
            for (node, timestep) in solution.keys():
                if node == i:
                    relevant_nodes.append(solution[(node, timestep)])
                if timestep == i:
                    relevant_timesteps.append(solution[(node, timestep)])
            if sum(relevant_nodes) != 1 or sum(relevant_timesteps) != 1:
                # timestep or nodes have more than 1 or 0 flags
                return None, end_time_measurement(start_time)

        # Check validity of solution
        if sum(value == 1 for value in solution.values()) > len(route):
            logging.warning("Result is longer than route! This might be problematic!")
            return None, end_time_measurement(start_time)

        # Run heuristic replacing None values
        if None in route:
            # get not assigned nodes
            nodes_unassigned = [node for node in list(nodes) if node not in route]
            nodes_unassigned = list(np.random.permutation(nodes_unassigned))
            for idx, node in enumerate(route):
                if node is None:
                    route[idx] = nodes_unassigned[0]
                    nodes_unassigned.remove(route[idx])

        # Cycle solution to start at provided start location
        if start is not None and route[0] != start:
            # Rotate to put the start in front
            idx = route.index(start)
            route = route[idx:] + route[:idx]

        # Log route
        parsed_route = ' ->\n'.join([f' Node {visit}' for visit in route])
        logging.info(f"Route found:\n{parsed_route}")

        return route, end_time_measurement(start_time)

    def validate(self, solution: list) -> tuple[bool, float]:
        """
        Checks if it is a valid TSP tour.

        :param solution: List containing the nodes of the solution
        :return: Boolean whether the solution is valid, time it took to validate
        """
        start = start_time_measurement()
        nodes = self.application.nodes()

        if solution is None:
            return False, end_time_measurement(start)
        elif len([node for node in list(nodes) if node not in solution]) == 0:
            logging.info(f"All {len(solution)} nodes got visited")
            return True, end_time_measurement(start)
        else:
            logging.error(f"{len([node for node in list(nodes) if node not in solution])} nodes were NOT visited")
            return False, end_time_measurement(start)

    def evaluate(self, solution: list) -> tuple[float, float]:
        """
        Find distance for given route and original data.

        :param solution: List containing the nodes of the solution
        :return: Tour cost and the time it took to calculate it
        """
        start = start_time_measurement()
        # Get the total distance without return
        total_dist = 0
        for idx, _ in enumerate(solution[:-1]):
            dist = self.application[solution[idx + 1]][solution[idx]]
            total_dist += dist['weight']

        logging.info(f"Total distance (without return): {total_dist}")

        # Add distance between start and end point to complete cycle
        return_distance = self.application[solution[0]][solution[-1]]['weight']

        # Get distance for full cycle
        distance_with_return = total_dist + return_distance
        logging.info(f"Total distance (including return): {distance_with_return}")

        return distance_with_return, end_time_measurement(start)

    def save(self, path: str, iter_count: int) -> None:
        """
        Save the current application state to a file.

        :param path: The directory path where the file will be saved
        :param iter_count: The iteration count to include in the filename
        """
        with open(f"{path}/graph_iter_{iter_count}.gpickle", "wb") as file:
            pickle.dump(self.application, file, pickle.HIGHEST_PROTOCOL)

    def visualize_solution(self, processed_solution: list[int], path: str):
        """
        Plot a graph representing the problem network with the solution path highlighted

        :param processed_solution: The solution already processed by :func:`process_solution`, a list of visited node IDs in order of being visited.
        :param path: File path for the plot
        :returns: None
        """
        NODE_SIZE = 300   # Default=300
        EDGE_WIDTH = 1.0  # Default=1.0
        FONT_SIZE = 12    # Default=12

        path_edges = list(nx.utils.pairwise(processed_solution, cyclic=True))
        path_edges = [(u, v) if u < v else (v, u) for (u, v) in path_edges]
        G = self.application
        pos = nx.circular_layout(G)
        weights = nx.get_edge_attributes(G, "weight")
        filtered_weights = {e: (int(weights[e])) for e in path_edges}

        nx.draw_networkx_nodes(G, pos, node_size=NODE_SIZE)
        nx.draw_networkx_edges(G, pos, edgelist=G.edges(), width=EDGE_WIDTH, edge_color="gray")
        nx.draw_networkx_edges(G, pos, edgelist=path_edges, width=2 * EDGE_WIDTH, edge_color="red", arrows=True)
        nx.draw_networkx_labels(G, pos, font_size=FONT_SIZE)
        nx.draw_networkx_edge_labels(G, pos, filtered_weights, font_size=.5 * FONT_SIZE)

        plt.savefig(path)
        plt.close()



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC


class Device(ABC):
    """
    The device class abstracts away details of the physical device,
    such as submitting a task to the quantum system.
    """

    def __init__(self, device_name: str):
        """
        Constructor method.

        :param device_name: Name of the device
        """
        self.device = None
        self.device_name = device_name
        self.config = None

    def get_parameter_options(self) -> dict:
        """
        Method to return the parameters to fine tune the device.

        Should always be in this format:

        .. code-block:: json

           {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How many reads do you want?"
               }
           }

        :return: Available device settings for this device
        """
        return {}

    def set_config(self, config: dict) -> None:
        """
        Sets the device configuration.

        :param config: Configuration dictionary
        """
        self.config = config

    def get_device(self) -> any:
        """
        Returns the device instance.

        :return: Instance of the device
        """
        return self.device

    def get_device_name(self) -> str:
        """
        Returns the name of the Device.

        :return: Name of the device
        """
        return self.device_name



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict

import networkx as nx
from networkx.algorithms import approximation as approx

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class ReverseGreedyClassicalTSP(Solver):
    """
    Classical Reverse Greedy Solver for the TSP.
    We take the worst choice at each step.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Local"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "networkx", "version": "3.4.2"}]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option == "Local":
            from quark.modules.devices.Local import Local  # pylint: disable=C0415
            return Local()
        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns empty dict as this solver has no configurable settings.

        :return: Empty dict
        """
        return {}

    class Config(TypedDict):
        """
        Empty config as this solver has no configurable settings.
        """
        pass

    def run(self, mapped_problem: nx.Graph, device_wrapper: any, config: Config, **kwargs: any) \
            -> tuple[dict, float, dict]:
        """
        Solve the TSP graph in a greedy fashion.

        :param mapped_problem: Graph representing a TSP
        :param device_wrapper: Local device
        :param config: Empty dict
        :param kwargs: No additionally settings needed
        :return: Solution, the time it took to compute it and optional additional information
        """
        # Need to deep copy since we are modifying the graph in this function.
        # Else the next repetition would work with a different graph
        mapped_problem = mapped_problem.copy()

        # Let's flip the edge weights to take the worst node every time instead of the best
        for _, _, d in mapped_problem.edges(data=True):
            d['weight'] = -1.0 * d['weight']

        start = start_time_measurement()

        tour = approx.greedy_tsp(mapped_problem)

        # We remove the duplicate node as we don't want a cycle
        # https://stackoverflow.com/a/7961390/10456906
        tour = list(dict.fromkeys(tour))

        # Parse tour so that it can be processed later
        result = {(node, idx): 1 for idx, node in enumerate(tour)}

        return result, end_time_measurement(start), {}



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod
from typing import final

from quark.utils import _get_instance_with_sub_options


class Application(ABC):
    """
    The application component defines the workload, comprising a dataset of increasing complexity,
    a validation, and an evaluation function.
    """

    def __init__(self, application_name: str):
        """
        Constructor method.

        :param application_name: Name of the application
        """
        self.application_name = application_name
        self.application = None
        self.mapping_options = []
        self.sub_options = []

        self.problem = None
        self.problems = {}
        self.conf_idx = None

        super().__init__()

    def get_application(self) -> any:
        """
        Getter that returns the application.

        :return: The application instance
        """
        return self.application

    @abstractmethod
    def get_solution_quality_unit(self) -> str:
        """
        Method to return the unit of the evaluation which is used to make the plots nicer.

        :return: String with the unit
        """

    @abstractmethod
    def get_parameter_options(self) -> dict:
        """
        Method to return the parameters needed to create a concrete problem of an application.

        Should always be in this format:

        .. code-block:: json

            {
               "parameter_name":{
                  "values":[1, 2, 3],
                  "description":"How many nodes do you need?"
               },
                "parameter_name_2":{
                  "values":["x", "y"],
                  "description":"Which type of problem do you want?"
               }
            }

        :return: Available application settings for this application
        """
        pass

    def regenerate_on_iteration(self, config: dict) -> bool:
        """
        Overwrite this to return True if the problem should be newly generated
        on every iteration. Typically, this will be the case if the problem is taken
        from a statistical ensemble e.g. an erdos-renyi graph.

        :param config: The application configuration
        :return: Whether the problem should be recreated on every iteration. Returns False if not overwritten.
        """
        return False

    @final
    def init_problem(self, config: dict, conf_idx: int, iter_count: int, path: str) -> any:
        """
        This method is called on every iteration and calls generate_problem if necessary.
        conf_idx identifies the application configuration.
        Note that when there are several mappings, solvers or devices this method is
        called several times with the same conf_idx and rep_count. In this case
        the problem will be the same if conf_idx and rep_count are both the same.

        The implementation assumes that the loop over the rep_count is within the loop
        over the different application configurations (conf_idx).

        :param config: the application configuration
        :param conf_idx: the index of the application configuration
        :param iter_count: the repetition count (starting with 1)
        :param path: the path used to save each newly generated problem instance
        :return: the current problem instance
        """
        if conf_idx != self.conf_idx:
            self.problems = {}
            self.conf_idx = conf_idx

        key = iter_count if self.regenerate_on_iteration(config) else "dummy"
        if key in self.problems:
            self.problem = self.problems[key]
        else:
            self.problem = self.generate_problem(config, iter_count)
            self.problems[key] = self.problem
            self.save(path, iter_count)
        return self.problem

    @abstractmethod
    def generate_problem(self, config: dict, iter_count: int) -> any:
        """
        Depending on the config this method creates a concrete problem and returns it.

        :param config: The application configuration
        :param iter_count: The iteration count
        :return: The generated problem instance
        """
        pass

    def process_solution(self, solution) -> tuple[any, float]:
        """
        Most of the time the solution has to be processed before it can be validated and evaluated
        This might not be necessary in all cases, so the default is to return the original solution.

        :param solution: The solution to be processed
        :return: Processed solution and the execution time to process it
        """
        return solution, 0

    @abstractmethod
    def validate(self, solution) -> tuple[bool, float]:
        """
        Check if the solution is valid.

        :param solution: The solution to validate
        :return: Boolean indicating if the solution is valid and the time it took to create it
        """
        pass

    @abstractmethod
    def evaluate(self, solution: any) -> tuple[float, float]:
        """
        Checks how good the solution is to allow comparison to other solutions.

        :param solution: The solution to evaluate
        :return: Evaluation and the time it took to create it
        """
        pass

    @abstractmethod
    def save(self, path: str, iter_count: int) -> None:
        """
        Save the concrete problem.

        :param path: Path of the experiment directory for this run
        :param iter_count: the iteration count
        """
        pass

    def get_submodule(self, mapping_option: str) -> any:
        """
        If self.sub_options is not None, a mapping is instantiated according to the information given in
        self.sub_options. Otherwise, get_mapping is called as fall back.

        :param mapping_option: The option for the mapping
        :return: instance of a mapping class
        """
        if self.sub_options is None:
            return self.get_mapping(mapping_option)
        else:
            return _get_instance_with_sub_options(self.sub_options, mapping_option)

    @abstractmethod
    def get_mapping(self, mapping_option: str) -> any:
        """
        Return a default mapping for an application. This applies only if
        self.sub_options is None. See get_submodule.

        :param mapping_option: String with the option
        :return: Instance of a mapping class
        """
        pass

    def get_available_mapping_options(self) -> list:
        """
        Gets the list of available mapping options.

        :return: list of mapping options
        """
        if self.sub_options is None:
            return self.mapping_options
        else:
            return [option["name"] for option in self.sub_options]



#  Copyright 2023 science + computing ag
#
#    Licensed under the Apache License, Version 2.0 (the "License");
#    you may not use this file except in compliance with the License.
#    You may obtain a copy of the License at
#
#        http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS,
#    WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#    See the License for the specific language governing permissions and
#    limitations under the License.
#

from abc import ABC
import json
import logging
from time import time

from quark.modules.Core import Core
from quark.modules.applications.Application import Application as Application_NEW
from quark.modules.applications.Mapping import Mapping as Mapping_NEW
from quark.modules.solvers.Solver import Solver as Solver_NEW
from quark.modules.devices.Device import Device as Device_NEW
from quark.quark2_adapter.legacy_classes.Application import Application as Application_OLD
from quark.quark2_adapter.legacy_classes.Mapping import Mapping as Mapping_OLD
from quark.quark2_adapter.legacy_classes.Solver import Solver as Solver_OLD
from quark.quark2_adapter.legacy_classes.Device import Device as Device_OLD


WARNING_MSG = 'Class "%s" is inheriting from deprecated base class. Please refactor your class.'


class ApplicationAdapter(Application_NEW, Application_OLD, ABC):
    """
    If you have a concrete Application written with an older QUARK version (before QUARK2) you can
    replace
    .. code-block:: python

        from applications.Application import Application

    by
    .. code-block:: python

        from quark2_adapter/adapters import ApplicationAdapter as Application

    to get your Application running with QUARK2.
    """

    def __init__(self, application_name: str, *args, **kwargs):
        """
        Constructor method.
        """
        logging.warning(WARNING_MSG, self.__class__.__name__)
        Application_NEW.__init__(self, application_name)
        Application_OLD.__init__(self, application_name)
        self.args = args
        self.kwargs = kwargs

        self.problem_conf_hash = None
        self.problems = {}

    @property
    def submodule_options(self) -> list[str]:
        """
        Maps the old attribute mapping_options to the new attribute submodule_options.
        """
        return self.mapping_options

    @submodule_options.setter
    def submodule_options(self, options: list[str]):
        """
        Maps the old attribute mapping_options to the new attribute submodule_options.

        :param options: List of submodule options
        """
        self.mapping_options = options

    def get_default_submodule(self, option: str) -> Core:
        """
        Maps the old method get_mapping to the new get_default_submodule.

        :param option: String with the chosen submodule
        :return: Module of type Core
        """
        return self.get_mapping(option)

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Application_NEW.preprocess using the Application_OLD interface.

        :param input_data: Data for the module, comes from the parent module if that exists
        :param config: Config for the module
        :param kwargs: Optional keyword arguments
        :return: The output of the preprocessing and the time it took to preprocess
        """
        start = time()
        logging.warning(WARNING_MSG, self.__class__.__name__)

        rep_count = kwargs["rep_count"]

        # create a hash value for identifying the problem configuration
        # compare https://stackoverflow.com/questions/5884066/hashing-a-dictionary
        problem_conf_hash = json.dumps(config, sort_keys=True)

        if self.problem_conf_hash != problem_conf_hash:
            self.problems = {}
            self.problem_conf_hash = problem_conf_hash

        problem_key = rep_count if self.regenerate_on_iteration(config) else "dummy"
        if problem_key in self.problems:
            self.problem, creation_time = self.problems[problem_key]
        else:
            start = time()
            logging.info("Generating new problem instance")
            self.problem = self.generate_problem(config, rep_count)
            creation_time = (time() - start) * 1000
            self.problems[problem_key] = (self.problem, creation_time)

        return self.problem, creation_time

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Application_NEW.postprocess using the Application_OLD interface.

        :param input_data: Input data comes from the submodule if that exists
        :param config: Config for the module
        :param kwargs: Optional keyword arguments
        :return: The output of the postprocessing and the time it took to postprocess
        """
        processed_solution, time_processing = self.process_solution(input_data)
        solution_validity, time_validate = self.validate(processed_solution)
        solution_quality, time_evaluate = (self.evaluate(processed_solution)
                                           if solution_validity else (None, 0.0))

        self.metrics.add_metric("time_to_validation", time_validate)
        self.metrics.add_metric("time_to_validation_unit", "ms")
        self.metrics.add_metric("solution_validity", solution_validity)
        self.metrics.add_metric("solution_quality", solution_quality)
        self.metrics.add_metric("solution_quality_unit", self.get_solution_quality_unit())

        return (solution_validity, solution_quality), time_validate + time_evaluate + time_processing


class MappingAdapter(Mapping_NEW, Mapping_OLD, ABC):
    """
    If you have a concrete Mapping written with an older QUARK version (before QUARK2) you can
    replace
    .. code-block:: python

        from applications.Mapping import Mapping

    by
    .. code-block:: python

        from quark2_adapter/adapters import MappingAdapter as Mapping

    to get your Mapping running with QUARK2.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor method.
        """
        Mapping_NEW.__init__(self)
        Mapping_OLD.__init__(self)

    @property
    def submodule_options(self) -> list[str]:
        """
        Maps the old attribute solver_options to the new attribute submodule_options.
        """
        return self.solver_options

    @submodule_options.setter
    def submodule_options(self, options: list[str]):
        """
        Maps the old attribute solver_options to the new attribute submodule_options.

        :param options: List of solver options
        """
        self.solver_options = options

    def get_default_submodule(self, option: str) -> Core:
        """
        Maps the old method get_solver to the new get_default_submodule.

        :param option: String with the chosen submodule
        :return: Module of type Core
        """
        return self.get_solver(option)

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Mapping_NEW.preprocess using the Mapping_OLD interface.
        """
        logging.warning(WARNING_MSG, self.__class__.__name__)
        return self.map(input_data, config=config)

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Mapping_NEW.postprocess using the Mapping_OLD interface.
        """
        logging.info("Calling %s.reverse_map", __class__.__name__)
        processed_solution, postprocessing_time = self.reverse_map(input_data)
        return processed_solution, postprocessing_time


def recursive_replace_dict_keys(obj: any) -> any:
    """
    Replace values used as dictionary keys by their string representation
    to make the object JSON-compatible.

    :param obj: The object to convert
    .return: The object with all dictionary keys converted to strings
    """
    obj_new = None
    if isinstance(obj, dict):
        obj_new = {str(key): recursive_replace_dict_keys(value) for key, value in obj.items()}
    elif isinstance(obj, list):
        obj_new = [recursive_replace_dict_keys(element) for element in obj]
    elif isinstance(obj, tuple):
        obj_new = tuple(recursive_replace_dict_keys(element) for element in obj)
    else:
        obj_new = obj

    return obj_new


class SolverAdapter(Solver_NEW, Solver_OLD, ABC):
    """
    If you have a concrete Solver written with an older QUARK version (before QUARK2) you can
    replace
    .. code-block:: python

        from solvers.Solver import Solver

    by
    .. code-block:: python

        from quark2_adapter/adapters import SolverAdapter as Solver

    to get your Solver running with QUARK2.
    """

    def __init__(self, *args, **kwargs):
        """
        Constructor method.
        """
        Solver_NEW.__init__(self)
        Solver_OLD.__init__(self)

    @property
    def submodule_options(self) -> list[str]:
        """
        Maps the old attribute device_options to the new attribute submodule_options.
        """
        return self.device_options

    @submodule_options.setter
    def submodule_options(self, options: list[str]):
        """
        Maps the old attribute device_options to the new attribute submodule_options.

        :param options: List of device options
        """
        self.device_options = options

    def get_default_submodule(self, option: str) -> Core:
        """
        Maps the old method get_device to the new get_default_submodule.

        :param option: String with the chosen submodule
        :return: Module of type Core
        """
        return self.get_device(option)

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Solver_NEW.preprocess using the Solver_OLD interface.

        :param input_data: Data for the module, comes from the parent module if that exists
        :param config: Config for the module
        :param kwargs: Optional keyword arguments
        :return: The output of the preprocessing and the time it took to preprocess
        """
        logging.warning(WARNING_MSG, self.__class__.__name__)
        return input_data, 0.0

    def postprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Solver_NEW.postprocess using the Solver_OLD interface.

        :param input_data: Data passed to the run function of the solver
        :param config: Solver config
        :param kwargs: Optional keyword arguments
        :return: Output and time needed
        """
        run_kwargs = {
            "store_dir": kwargs["store_dir"],
            "repetition": kwargs["rep_count"]
        }
        raw_solution, runtime, additional_solver_information = self.run(
            input_data["mapped_problem"],
            device_wrapper=input_data["device"],
            config=config,
            **run_kwargs
        )

        self.metrics.add_metric(
            "additional_solver_information", dict(additional_solver_information)
        )
        self.metrics.add_metric(
            "solution_raw", self.raw_solution_to_json(raw_solution)
        )

        return raw_solution, runtime

    def raw_solution_to_json(self, raw_solution: any) -> any:
        """
        In case the raw_solution provided by your concrete solver is not directly representable as json
        you can overwrite SolverAdapter.raw_solution_to_json in your Solver to implement the conversion
        to json.
        Note that using 'recursive_replace_dict_keys' provided by this module might help.

        :param raw_solution: The raw solution
        :return: JSON-compatible representation of the raw solution
        """
        return raw_solution


class DeviceAdapter(Device_NEW, Device_OLD):
    """
    If you have a concrete Device written with an older QUARK version (before QUARK2) you can
    replace
    .. code-block:: python

        from devices.Device import Device

    by
    .. code-block:: python

        from quark2_adapter/adapters import DeviceAdapter as Device

    to get your Device running with QUARK2.
    """

    def __init__(self, name: str):
        """
        Constructor method.
        """
        Device_NEW.__init__(self, name)
        Device_OLD.__init__(self, name)
        self.device_name = name

    def get_default_submodule(self, option: str) -> Core:
        """
        Implements get_default_submodule by returning None as before QUARK2 a Device
        could not have submodules.

        :param option: String with the chosen submodule
        :return: None
        """
        return None

    def preprocess(self, input_data: any, config: dict, **kwargs) -> tuple[any, float]:
        """
        Implements Device_NEW.preprocess using the Device_OLD interface.

        :param input_data: Data for the module, comes from the parent module if that exists
        :param config: Config for the device
        :param kwargs: Optional keyword arguments
        :return: The output of the preprocessing and the time it took to preprocess
        """
        logging.warning(WARNING_MSG, self.__class__.__name__)
        self.set_config(config)
        return {"mapped_problem": input_data, "device": self}, 0.


class LocalAdapter(DeviceAdapter):
    """
    If you have been using the device devices.Local.Local from an older QUARK version (before QUARK2)
    you can replace
    .. code-block:: python

        from devices.Local import Local

    by
    .. code-block:: python

        from quark2_adapter/adapters import LocalAdapter as Local

    to get your code running with QUARK2.
    """

    def __init__(self):
        """
        Constructor method.
        """
        DeviceAdapter.__init__(self, name="local")
        self.device = None



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict, Union
import logging

import numpy as np
import pkg_resources

from quark.utils import start_time_measurement, end_time_measurement
from quark.modules.applications.qml.generative_modeling.transformations.MinMax import MinMax
from quark.modules.applications.qml.generative_modeling.transformations.PIT import PIT
from quark.modules.applications.qml.generative_modeling.data.data_handler.DataHandlerGenerative import DataHandlerGenerative


class ContinuousData(DataHandlerGenerative):
    """
    A data handler for continuous datasets. This class loads a dataset from a specified path and provides
    methods for data transformation and evaluation.

    """

    def __init__(self):
        """
        The continuous data class loads a dataset from the path
        src/modules/applications/qml/generative_modeling/data
        """
        super().__init__("")
        self.submodule_options = ["PIT", "MinMax"]
        self.transformation = None
        self.dataset = None
        self.n_registers = None
        self.gc = None
        self.n_qubits = None

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "numpy", "version": "1.26.4"}]

    def get_default_submodule(self, option: str) -> Union[PIT, MinMax]:
        if option == "MinMax":
            self.transformation = MinMax()
        elif option == "PIT":
            self.transformation = PIT()
        else:
            raise NotImplementedError(f"Transformation Option {option} not implemented")
        return self.transformation

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this application.

        :return: Dictionary of parameter options
        .. code-block:: python

            return {
                "data_set": {
                    "values": ["X_2D", "O_2D", "MG_2D", "Stocks_2D"],
                    "description": "Which dataset do you want to use?"
                },

                "train_size": {
                    "values": [0.1, 0.3, 0.5, 0.7, 1.0],
                    "description": "What percentage of the dataset do you want to use for training?"
                }
            }
        """
        return {
            "data_set": {
                "values": ["X_2D", "O_2D", "MG_2D", "Stocks_2D"],
                "description": "Which dataset do you want to use?"
            },

            "train_size": {
                "values": [0.1, 0.3, 0.5, 0.7, 1.0],
                "description": "What percentage of the dataset do you want to use for training?"
            }

        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            data_set: str

        """

        data_set: int
        train_size: float

    def data_load(self, gen_mod: dict, config: Config) -> dict:
        """
        The chosen dataset is loaded and split into a training set.

        :param gen_mod: Dictionary with collected information of the previous modules
        :param config: Config specifying the parameters of the data handler
        :return: Dictionary including the mapped problem
        """
        self.dataset_name = config["data_set"]
        self.n_qubits = gen_mod["n_qubits"]

        filename = pkg_resources.resource_filename(
            'modules.applications.qml.generative_modeling.data',
            f"{self.dataset_name}.npy"
        )
        self.dataset = np.load(filename)

        application_config = {
            "dataset_name": self.dataset_name,
            "n_qubits": self.n_qubits,
            "dataset": self.dataset,
            "train_size": config["train_size"],
            "store_dir_iter": gen_mod["store_dir_iter"]}

        return application_config

    def evaluate(self, solution: dict) -> tuple[float, float]:
        """
        Calculates KL in original space.

        :param solution: a dictionary containing the solution data, including histogram_generated_original
                         and histogram_train_original
        :return: Kullback-Leibler (KL) divergence for the generated samples and the time it took to calculate it
        """
        start = start_time_measurement()

        generated = solution["histogram_generated_original"]
        generated[generated == 0] = 1e-8

        histogram_train_original = solution["histogram_train_original"]
        histogram_train_original[histogram_train_original == 0] = 1e-8

        target = histogram_train_original.ravel()
        generated = generated.ravel()
        kl_divergence = self.kl_divergence(target, generated)

        logging.info(f"KL original space: {kl_divergence}")

        return kl_divergence, end_time_measurement(start)

    def kl_divergence(self, target: np.ndarray, q: np.ndarray) -> float:
        """
        Function to calculate KL divergence.

        :param target: Probability mass function of the target distribution
        :param q: Probability mass function generated by the quantum circuit
        :return: Kullback-Leibler divergence
        """
        if q.shape != target.shape:
            q = np.resize(q, target.shape)

        return np.sum(target * np.log(target / q))



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from abc import ABC, abstractmethod


class Training(ABC):
    """
    Abstract base class for training QML models.
    """

    @abstractmethod
    def start_training(self, input_data: dict, config: any, **kwargs: dict) -> dict:
        """
        This function starts the training of QML model or deploys a pretrained model.

        :param input_data: A representation of the quantum machine learning model that will be trained
        :param config: Config specifying the parameters of the training (dict-like Config type defined in children)
        :param kwargs: Optional additional settings
        :return: Solution, the time it took to compute it and some optional additional information
        """
        pass



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import Union

import numpy as np
import pennylane as qml
from jax import numpy as jnp
import jax

from quark.modules.applications.qml.generative_modeling.mappings.LibraryGenerative import LibraryGenerative
from quark.modules.applications.qml.generative_modeling.training.Inference import Inference
from quark.modules.applications.qml.generative_modeling.training.QGAN import QGAN
from quark.modules.applications.qml.generative_modeling.training.QCBM import QCBM

jax.config.update("jax_enable_x64", True)


class LibraryPennylane(LibraryGenerative):

    def __init__(self):
        super().__init__("LibraryPennylane")
        self.submodule_options = ["QCBM", "QGAN", "Inference"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Returns requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [
            {"name": "pennylane", "version": "0.39.0"},
            {"name": "pennylane-lightning", "version": "0.39.0"},
            {"name": "numpy", "version": "1.26.4"},
            {"name": "jax", "version": "0.4.35"},
            {"name": "jaxlib", "version": "0.4.35"}
        ]

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for the PennyLane Library.

        :return: Dictionary with configurable settings.
        .. code-block:: python

            return {
                "backend": {
                    "values": ["default.qubit", "default.qubit.jax", "lightning.qubit", "lightning.gpu"],
                    "description": "Which backend do you want to use?"
                },

                "n_shots": {
                    "values": [100, 1000, 10000, 1000000],
                    "description": "How many shots do you want use for estimating the PMF of the model?"
                }
            }
        """
        return {
            "backend": {
                "values": ["default.qubit", "default.qubit.jax", "lightning.qubit", "lightning.gpu"],
                "description": "Which device do you want to use?"
            },
            "n_shots": {
                "values": [100, 1000, 10000, 1000000],
                "description": "How many shots do you want use for estimating the PMF of the model?"
            }
        }

    def get_default_submodule(self, option: str) -> Union[QCBM, QGAN, Inference]:
        """
        Returns the default submodule based on the provided option.

        :param option: The option to select the submodule
        :return: The selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """

        if option == "QCBM":
            return QCBM()
        elif option == "QGAN":
            return QGAN()
        elif option == "Inference":
            return Inference()
        else:
            raise NotImplementedError(f"Training option {option} not implemented")

    def sequence_to_circuit(self, input_data: dict) -> dict:
        """
        Method that maps the gate sequence, that specifies the architecture of a quantum circuit
        to its PennyLane implementation.

        :param input_data: Collected information of the benchmarking process
        :return: Same dictionary but the gate sequence is replaced by its PennyLane implementation
        """
        gate_sequence = input_data["gate_sequence"]
        n_qubits = input_data["n_qubits"]
        num_parameters = sum(
            1 for gate, _ in gate_sequence if gate in ["RZ", "RX", "RY", "RXX", "RYY", "RZZ", "CRY"]
        )

        def create_circuit(params):
            param_counter = 0
            for gate, wires in gate_sequence:
                if gate == 'Hadamard':
                    qml.Hadamard(wires[0])
                elif gate == 'CNOT':
                    qml.CNOT(wires)
                elif gate == 'RZ':
                    qml.RZ(params[param_counter], wires[0])
                    param_counter += 1
                elif gate == 'RX':
                    qml.RX(params[param_counter], wires[0])
                    param_counter += 1
                elif gate == 'RY':
                    qml.RY(params[param_counter], wires[0])
                    param_counter += 1
                elif gate == "RXX":
                    qml.IsingXX(params[param_counter], wires)
                    param_counter += 1
                elif gate == "RYY":
                    qml.IsingYY(params[param_counter], wires)
                    param_counter += 1
                elif gate == "RZZ":
                    qml.IsingZZ(params[param_counter], wires)
                    param_counter += 1
                elif gate == "CRY":
                    qml.CRY(params[param_counter], wires)
                    param_counter += 1
                elif gate == "Barrier":
                    qml.Barrier()
                elif gate == 'Measure':
                    continue
                else:
                    raise NotImplementedError(f"Gate {gate} not implemented")

            return qml.probs(wires=list(range(n_qubits)))

        input_data["n_params"] = num_parameters
        input_data.pop("gate_sequence")
        input_data["circuit"] = create_circuit

        return input_data

    @staticmethod
    def select_backend(config: str, n_qubits: int) -> any:
        """
        This method configures the backend.

        :param config: Name of a backend
        :param n_qubits: Number of qubits
        :return: Configured backend
        """
        if config == "lightning.gpu":
            backend = qml.device(name="lightning.gpu", wires=n_qubits)
        elif config == "lightning.qubit":
            backend = qml.device(name="lightning.qubit", wires=n_qubits)
        elif config == "default.qubit":
            backend = qml.device(name="default.qubit", wires=n_qubits)
        elif config == "default.qubit.jax":
            backend = qml.device(name="default.qubit.jax", wires=n_qubits)
        else:
            raise NotImplementedError(f"Device Configuration {config} not implemented")

        return backend

    @staticmethod
    def get_execute_circuit(circuit: callable, backend: qml.device, config: str, config_dict: dict) \
            -> tuple[any, any]:
        """
        This method combines the PennyLane circuit implementation and the selected backend and returns a function
        that will be called during training.

        :param circuit: PennyLane implementation of the quantum circuit
        :param backend: Configured qiskit backend
        :param config: Name of a backend
        :param config_dict: Dictionary including the number of shots
        :return: Tuple that contains a method that executes the quantum circuit for a given set of parameters twice
        """
        n_shots = config_dict["n_shots"]

        if config == "default.qubit.jax":
            qnode = qml.QNode(circuit, device=backend, interface="jax")
            qnode = jax.jit(qnode)
        else:
            qnode = qml.QNode(circuit, device=backend)

        def execute_circuit(solutions):
            if config == "default.qubit.jax":
                solutions = jnp.asarray(solutions)

            pmfs, samples = [], []
            for solution in solutions:
                pmf = qnode(solution)
                pmfs.append(pmf)
                samples.append(np.random.multinomial(n_shots, pmf))

            return np.asarray(pmfs), np.asarray(samples)

        return execute_circuit, execute_circuit



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import glob
import json
import logging
import os
import os.path
from copy import deepcopy
from datetime import datetime
from enum import Enum
from pathlib import Path
from typing import Optional

import numpy as np

from quark.ConfigManager import ConfigManager
from quark.BenchmarkRecord import BenchmarkRecord, BenchmarkRecordStored
from quark.Plotter import Plotter
from quark.modules.Core import Core
from quark.utils import get_git_revision
from quark.utils_mpi import get_comm

comm = get_comm()


class Instruction(Enum):
    PROCEED = 0
    INTERRUPT = 1


class JobStatus(Enum):
    UNDEF = 0
    INTERRUPTED = 1
    FINISHED = 2
    FAILED = 3


def _prepend_instruction(result: tuple) -> tuple[Instruction, tuple]:
    """
    If the given list does not contain an Instruction as first entry a
    PROCEED is inserted at position 0 such that it is guaranteed that
    the first entry of the returned list is an Instruction with PROCEED
    as default.

    :param result: The tuple to which the Instruction is to be prepended
    :return: The tuple with an Instruction as first entry
    """
    if isinstance(result[0], Instruction):
        return result
    else:
        return Instruction.PROCEED, *result


def postprocess(module_instance: Core, *args, **kwargs) -> tuple[Instruction, tuple]:
    """
    Wraps module_instance.postprocess such that the first entry of the
    result list is guaranteed to be an Instruction. See _prepend_instruction.

    :param module_instance: The QUARK module on which to call postprocess
    :return: The result list of module_instance.postprocess with an Instruction as first entry.
    """
    result = module_instance.postprocess(*args, **kwargs)
    return _prepend_instruction(result)


def preprocess(module_instance: Core, *args, **kwargs) -> tuple[Instruction, tuple]:
    """
    Wraps module_instance.preprocess such that the first entry of the
    result list is guaranteed to be an Instruction. See _prepend_instruction.

    :param module_instance: The QUARK module on which to call preprocess
    :return: The result list of module_instance.preprocess with an Instruction as first entry.
    """
    result = module_instance.preprocess(*args, **kwargs)
    return _prepend_instruction(result)


class BenchmarkManager:
    """
    The benchmark manager is the main component of QUARK, orchestrating the overall benchmarking process.
    Based on the configuration, the benchmark manager will create an experimental plan considering all combinations of
    configurations, e.g., different problem sizes, solver, and hardware combinations. It will then instantiate the
    respective framework components. After executing the benchmarks, it collects the generated data and saves it.
    """

    def __init__(self, fail_fast: bool = False):
        """
        Constructor method.

        :param fail_fast: Boolean whether a single failed benchmark run causes QUARK to fail
        """
        self.fail_fast = fail_fast
        self.application = None
        self.application_configs = None  # TODO Seems to be unused, maybe delete
        self.results = []
        self.store_dir = None
        self.benchmark_record_template = None
        self.interrupted_results_path = None

    def load_interrupted_results(self) -> Optional[list]:
        """
        Loads the interrupted results if available.

        :return: The content of the results file from the QUARK run to be resumed or None.
        """
        if self.interrupted_results_path is None or not os.path.exists(self.interrupted_results_path):
            return None
        with open(self.interrupted_results_path, encoding='utf-8') as results_file:
            results = json.load(results_file)
        return results

    def _create_store_dir(self, store_dir: str = None, tag: str = None) -> None:
        """
        Creates directory for a benchmark run.

        :param store_dir: Directory where the new directory should be created
        :param tag: Prefix of the new directory
        """
        if store_dir is None:
            store_dir = Path.cwd()
        self.store_dir = f"{store_dir}/benchmark_runs/{tag + '-' if not None else ''}" \
            f"{datetime.today().strftime('%Y-%m-%d-%H-%M-%S')}"
        Path(self.store_dir).mkdir(parents=True, exist_ok=True)
        self._set_logger()

    def _resume_store_dir(self, store_dir: str) -> None:
        """
        Resumes the existing store directory.

        :param store-dir: Directory to be resumed
        """
        self.store_dir = store_dir
        self._set_logger()

    def _set_logger(self) -> None:
        """
        Sets up the logger to also write to a file in the store directory.
        """
        logger = logging.getLogger()
        formatter = logging.Formatter("%(asctime)s [%(levelname)s] %(message)s")
        filehandler = logging.FileHandler(f"{self.store_dir}/logging.log")
        filehandler.setFormatter(formatter)
        logger.addHandler(filehandler)

    def orchestrate_benchmark(self, benchmark_config_manager: ConfigManager, app_modules: list[dict],
                              store_dir: str = None, interrupted_results_path: str = None) -> None:
        """
        Executes the benchmarks according to the given settings.

        :param benchmark_config_manager: Instance of BenchmarkConfigManager class, where config is already set
        :param app_modules: The list of application modules as specified in the application modules configuration
        :param store_dir: Target directory to store the results of the benchmark (if user decided to store it)
        :param interrupted_results_path: Result file from which the information for the interrupted jobs will be read.
                                         If store_dir is None the parent directory of interrupted_results_path will
                                         be used as store_dir.
        """
        self.interrupted_results_path = interrupted_results_path
        if interrupted_results_path and not store_dir:
            self._resume_store_dir(os.path.dirname(interrupted_results_path))
        else:
            self._create_store_dir(store_dir, tag=benchmark_config_manager.get_config()["application"]["name"].lower())

        benchmark_config_manager.save(self.store_dir)
        benchmark_config_manager.load_config(app_modules)
        self.application = benchmark_config_manager.get_app()
        benchmark_config_manager.create_tree_figure(self.store_dir)

        logging.info(f"Created Benchmark run directory {self.store_dir}")

        benchmark_backlog = benchmark_config_manager.start_create_benchmark_backlog()
        self.run_benchmark(benchmark_backlog, benchmark_config_manager.get_reps())

        # Wait until all MPI processes have finished and save results on rank 0
        comm.Barrier()
        if comm.Get_rank() == 0:
            results = self._collect_all_results()
            self._save_as_json(results)

    def run_benchmark(self, benchmark_backlog: list, repetitions: int) -> None:  # pylint: disable=R0915
        """
        Goes through the benchmark backlog, which contains all the benchmarks to execute.

        :param repetitions: Number of repetitions
        :param benchmark_backlog: List with the benchmark items to run
        """
        git_dir = os.path.join(os.path.dirname(os.path.abspath(__file__)), "..", )
        git_revision_number, git_uncommitted_changes = get_git_revision(git_dir)
        break_flag = False

        job_status_count_total = {}
        interrupted_results = self.load_interrupted_results()
        for idx_backlog, backlog_item in enumerate(benchmark_backlog):
            benchmark_records: list[BenchmarkRecord] = []
            path = f"{self.store_dir}/benchmark_{idx_backlog}"
            Path(path).mkdir(parents=True, exist_ok=True)
            with open(f"{path}/application_config.json", 'w') as filehandler:
                json.dump(backlog_item["config"], filehandler, indent=2)
            job_status_count = {}

            for i in range(1, repetitions + 1):
                logging.info(f"Running backlog item {idx_backlog + 1}/{len(benchmark_backlog)},"
                             f" Iteration {i}/{repetitions}:")

                # Getting information of interrupted jobs
                job_info_with_meta_data = {}
                if interrupted_results:
                    for entry in interrupted_results:
                        if entry["benchmark_backlog_item_number"] == idx_backlog and entry["repetition"] == i:
                            job_info_with_meta_data = entry
                            break
                job_info = job_info_with_meta_data['module'] if job_info_with_meta_data else {}
                quark_job_status_name = job_info.get("quark_job_status")

                if quark_job_status_name in (JobStatus.FINISHED.name, JobStatus.FAILED.name):
                    quark_job_status = JobStatus.FINISHED if quark_job_status_name == JobStatus.FINISHED.name \
                        else JobStatus.FAILED
                    benchmark_records.append(BenchmarkRecordStored(job_info_with_meta_data))
                    job_status_count[quark_job_status] = job_status_count.get(quark_job_status, 0) + 1
                    job_status_count_total[quark_job_status] = job_status_count_total.get(quark_job_status, 0) + 1
                    logging.info("job already %s - skip.", quark_job_status_name)
                    continue

                try:
                    self.benchmark_record_template = BenchmarkRecord(
                        idx_backlog,
                        datetime.today().strftime('%Y-%m-%d-%H-%M-%S'),
                        git_revision_number, git_uncommitted_changes,
                        i, repetitions
                    )
                    self.application.metrics.set_module_config(backlog_item["config"])
                    instruction, problem, preprocessing_time = preprocess(
                        self.application, None, backlog_item["config"],
                        store_dir=path, rep_count=i, previous_job_info=job_info
                    )
                    self.application.metrics.set_preprocessing_time(preprocessing_time)
                    self.application.save(path, i)

                    postprocessing_time = 0.0
                    benchmark_record = self.benchmark_record_template.copy()
                    if instruction == Instruction.PROCEED:
                        instruction, processed_input, benchmark_record = \
                            self.traverse_config(backlog_item["submodule"], problem,
                                                 path, rep_count=i, previous_job_info=job_info)
                        if instruction == Instruction.PROCEED:
                            instruction, _, postprocessing_time = \
                                postprocess(self.application, processed_input, backlog_item["config"],
                                            store_dir=path, rep_count=i, previous_job_info=job_info)

                    if instruction == Instruction.INTERRUPT:
                        quark_job_status = JobStatus.INTERRUPTED
                    else:
                        quark_job_status = JobStatus.FINISHED

                    self.application.metrics.add_metric("quark_job_status", quark_job_status.name)
                    self.application.metrics.set_postprocessing_time(postprocessing_time)
                    self.application.metrics.validate()

                    if benchmark_record is not None:
                        benchmark_record.append_module_record_left(deepcopy(self.application.metrics))
                        benchmark_records.append(benchmark_record)

                except KeyboardInterrupt:
                    logging.warning("CTRL-C detected during run_benchmark. Still trying to create results.json.")
                    break_flag = True
                    break

                except Exception as error:
                    logging.exception(f"Error during benchmark run: {error}", exc_info=True)
                    quark_job_status = JobStatus.FAILED
                    if job_info:
                        # Restore results/infos from previous run
                        benchmark_records.append(job_info)
                    if self.fail_fast:
                        raise

                job_status_count[quark_job_status] = job_status_count.get(quark_job_status, 0) + 1
                job_status_count_total[quark_job_status] = job_status_count_total.get(quark_job_status, 0) + 1

            for record in benchmark_records:
                record.sum_up_times()

            for record in benchmark_records:
                record.sum_up_times()

            status_report = " ".join([f"{status.name}:{count}" for status, count in job_status_count.items()])
            logging.info("")
            logging.info(f" ==== Run backlog item {idx_backlog + 1}/{len(benchmark_backlog)} "
                         f"with {repetitions} iterations - {status_report} ==== ")
            logging.info("")

            # Wait until all MPI processes have finished and save results on rank 0
            comm.Barrier()
            if comm.Get_rank() == 0:
                with open(f"{path}/results.json", 'w') as filehandler:
                    json.dump([x.get() for x in benchmark_records], filehandler, indent=2, cls=NumpyEncoder)

            logging.info("")
            logging.info(" =============== Run finished =============== ")
            logging.info("")

            if break_flag:
                break

        # Log overall status information
        status_report = " ".join([f"{status.name}:{count}" for status, count in job_status_count_total.items()])
        logging.info(80 * "=")
        logging.info(f"====== Run {len(benchmark_backlog)} backlog items "
                     f"with {repetitions} iterations - {status_report}")

        nb_interrupted = job_status_count_total.get(JobStatus.INTERRUPTED, 0)
        nb_not_started = sum(job_status_count_total.values()) < len(benchmark_backlog)
        if nb_interrupted + nb_not_started > 0:
            try:
                rel_path = Path(self.store_dir).relative_to(os.getcwd())
            except ValueError:
                rel_path = self.store_dir
            logging.info("====== There are interrupted jobs. You may resume them by running QUARK with")
            logging.info(f"====== --resume-dir={rel_path}")
        logging.info(80 * "=")
        logging.info("")

    # pylint: disable=R0917
    def traverse_config(self, module: dict, input_data: any, path: str, rep_count: int, previous_job_info:
                        dict = None) -> tuple[Instruction, any, BenchmarkRecord]:
        """
        Executes a benchmark by traversing down the initialized config recursively until it reaches the end. Then
        traverses up again. Once it reaches the root/application, a benchmark run is finished.

        :param module: Current module
        :param input_data: The input data needed to execute the current module
        :param path: Path in case the modules want to store anything
        :param rep_count: The iteration count
        :param previous_job_info: Information about previous job
        :return: Tuple with the output of this step and the according BenchmarkRecord
        """
        # Only the value of the dict is needed (dict has only one key)
        module = module[next(iter(module))]
        module_instance: Core = module["instance"]

        submodule_job_info = None
        if previous_job_info and previous_job_info.get("submodule"):
            assert module['name'] == previous_job_info["submodule"]["module_name"], \
                f"asyncronous job info given, but no information about module {module['name']} stored in it"  # TODO
            if 'submodule' in previous_job_info and previous_job_info['submodule']:
                submodule_job_info = previous_job_info['submodule']

        module_instance.metrics.set_module_config(module["config"])
        instruction, module_instance.preprocessed_input, preprocessing_time = preprocess(
            module_instance, input_data,
            module["config"], store_dir=path,
            rep_count=rep_count,
            previous_job_info=submodule_job_info
        )

        module_instance.metrics.set_preprocessing_time(preprocessing_time)
        output = None
        benchmark_record = self.benchmark_record_template.copy()
        postprocessing_time = 0.0

        if instruction == Instruction.PROCEED:
            # Check if end of the chain is reached
            if not module["submodule"]:
                # If we reach the end of the chain we create the benchmark record, fill it and then pass it up
                instruction, module_instance.postprocessed_input, postprocessing_time = postprocess(
                    module_instance,
                    module_instance.preprocessed_input,
                    module["config"], store_dir=path,
                    rep_count=rep_count,
                    previous_job_info=submodule_job_info
                )
                output = module_instance.postprocessed_input
            else:
                instruction, processed_input, benchmark_record = self.traverse_config(
                    module["submodule"],
                    module_instance.preprocessed_input, path,
                    rep_count, previous_job_info=submodule_job_info
                )

                if instruction == Instruction.PROCEED:
                    instruction, module_instance.postprocessed_input, postprocessing_time = postprocess(
                        module_instance, processed_input,
                        module["config"], store_dir=path,
                        rep_count=rep_count,
                        previous_job_info=submodule_job_info
                    )
                    output = module_instance.postprocessed_input
                else:
                    output = processed_input

        module_instance.metrics.set_postprocessing_time(postprocessing_time)
        module_instance.metrics.validate()
        benchmark_record.append_module_record_left(deepcopy(module_instance.metrics))

        return instruction, output, benchmark_record

    def _collect_all_results(self) -> list[dict]:
        """
        Collect all results from the multiple results.json.

        :return: List of dicts with results
        """
        results = []
        for filename in glob.glob(f"{self.store_dir}/**/results.json"):
            with open(filename) as f:
                results += json.load(f)

        if len(results) == 0:
            logging.error("No results.json files were found! Probably an error occurred during execution.")
        return results

    def _save_as_json(self, results: list) -> None:
        """
        Saves benchmark results to a JSON file.

        :param results: Benchmark results to be saved
        """
        logging.info(f"Saving {len(results)} benchmark records to {self.store_dir}/results.json")
        with open(f"{self.store_dir}/results.json", 'w') as filehandler:
            json.dump(results, filehandler, indent=2)

    def summarize_results(self, input_dirs: list) -> None:
        """
        Helper function to summarize multiple experiments.

        :param input_dirs: List of directories
        """
        self._create_store_dir(tag="summary")
        logging.info(f"Summarizing {len(input_dirs)} benchmark directories")
        results = self.load_results(input_dirs)
        self._save_as_json(results)
        Plotter.visualize_results(results, self.store_dir)

    def load_results(self, input_dirs: list = None) -> list:
        """
        Load results from one or more results.json files.

        :param input_dirs: If you want to load more than 1 results.json (default is just 1, the one from the experiment)
        :return: A list
        """
        if input_dirs is None:
            input_dirs = [self.store_dir]

        results = []
        for input_dir in input_dirs:
            for filename in glob.glob(f"{input_dir}/results.json"):
                with open(filename) as f:
                    results += json.load(f)

        return results


class NumpyEncoder(json.JSONEncoder):
    """
    Encoder that is used for json.dump(...) since numpy value items in dictionary might cause problems.
    """

    def default(self, o: any):
        if isinstance(o, np.integer):
            return int(o)
        if isinstance(o, np.floating):
            return float(o)
        if isinstance(o, np.bool_):
            return bool(o)
        if isinstance(o, np.ndarray):
            return o.tolist()
        return super().default(o)



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import final


class Metrics:
    """
    Metrics Module, used by every QUARK module.
    """

    def __init__(self, module_name: str, module_src: str):
        """
        Constructor for Metrics class.

        :param module_name: Name of the module this metrics object belongs to
        :param module_src: Source file of the module this metrics object belongs to
        """
        self.module_name = module_name
        self.module_src = module_src
        self.preprocessing_time = None
        self.preprocessing_time_unit = "ms"
        self.postprocessing_time = None
        self.module_config = None
        self.total_time = None
        self.total_time_unit = "ms"
        self.postprocessing_time_unit = "ms"
        self.additional_metrics = {}

    @final
    def validate(self) -> None:
        """
        Validates whether the mandatory metrics got recorded, then sets total time.
        """
        assert self.preprocessing_time is not None, (
            "preprocessing time must not be None!"
        )
        assert self.postprocessing_time is not None, (
            "postprocessing time must not be None!"
        )
        self.total_time = self.preprocessing_time + self.postprocessing_time

    @final
    def set_preprocessing_time(self, value: float) -> None:
        """
        Sets the preprocessing time.

        :param value: Time
        """
        self.preprocessing_time = value

    @final
    def set_module_config(self, config: dict) -> None:
        """
        Sets the config of the module this metrics object belongs to.

        :param config: Config of the QUARK module
        """
        self.module_config = config

    @final
    def set_postprocessing_time(self, value: float) -> None:
        """
        Sets the postprocessing time.

        :param value: Time
        """
        self.postprocessing_time = value

    @final
    def add_metric(self, name: str, value: any) -> None:
        """
        Adds a single metric.

        :param name: Name of the metric
        :param value: Value of the metric
        """
        self.additional_metrics.update({name: value})

    @final
    def add_metric_batch(self, key_values: dict) -> None:
        """
        Adds a dictionary containing metrics to the existing metrics.

        :param key_values: Dict containing metrics
        """
        self.additional_metrics.update(key_values)

    @final
    def reset(self) -> None:
        """
        Resets all recorded metrics.
        """
        self.preprocessing_time = None
        self.postprocessing_time = None
        self.additional_metrics = {}

    @final
    def get(self) -> dict:
        """
        Returns all recorded metrics.

        :return: Metrics as a dict
        """
        return {
            "module_name": self.module_name,
            "module_src": self.module_src,
            "module_config": self.module_config,
            "total_time": self.total_time,
            "total_time_unit": self.total_time_unit,
            "preprocessing_time": self.preprocessing_time,
            "preprocessing_time_unit": self.preprocessing_time_unit,
            "postprocessing_time": self.postprocessing_time,
            "postprocessing_time_unit": self.postprocessing_time_unit,
            **self.additional_metrics
        }



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

import logging
from typing import TypedDict

from qrisp import QuantumVariable
from qrisp.algorithms.qiro import (
    QIROProblem,
    create_max_indep_replacement_routine,
    create_max_indep_cost_operator_reduced,
    qiro_rx_mixer,
    qiro_init_function
)
from qrisp.qaoa import create_max_indep_set_cl_cost_function

from quark.modules.solvers.Solver import Solver, Core
from quark.utils import start_time_measurement, end_time_measurement


class QIROSolver(Solver):
    """
    Run the QIRO implementation within the Qrisp local simulator backend.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["qrisp_simulator"]

    @staticmethod
    def get_requirements() -> list[dict]:
        """
        Return requirements of this module.

        :return: List of dict with requirements of this module
        """
        return [{"name": "qrisp", "version": "0.5.2"}]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the given option.

        :param option: The submodule option to select
        :return: Instance of the selected submodule
        :raises NotImplemented: If the provided option is not implemented
        """
        if option == "qrisp_simulator":
            from quark.modules.devices.qrisp_simulator.QrispSimulator import QrispSimulator  # pylint: disable=C0415
            return QrispSimulator()  # pylint: disable=E1102

        else:
            raise NotImplementedError(f"Device Option {option} not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return:
        .. code-block:: python

                    return {
                            "shots": {  # number measurements to make on circuit
                                "values": [10, 500, 1000, 2000, 5000, 10000],
                                "description": "How many shots do you need?"
                            },
                            "iterations": {  # number of optimization iterations
                                "values": [5, 10, 20, 50, 75],
                                "description": "How many optimization iterations do you need?"
                            },
                            "depth": { # depth of original QAOA
                                "values": [2, 3, 4, 5, 10],
                                "description": "How many layers for QAOA (Parameter: p) do you want?"
                            },
                            "QIRO_reps": { # number of QIRO reps
                                "values": [2, 3, 4, 5],
                                "description": "How QIRO reps (Parameter: n) do you want? Choose this
                                                "parameter sensibly in relation to the graph size!"
                            }
                        }
        """

        return {
            "shots": {  # number measurements to make on circuit
                "values": [10, 500, 1000, 2000, 5000, 10000],
                "description": "How many shots do you need?"
            },
            "iterations": {  # number of optimization iterations
                "values": [5, 10, 20, 50, 75],
                "description": "How many optimization iterations do you need?"
            },
            "depth": {  # depth of original QAOA
                "values": [2, 3, 4, 5, 10],
                "description": "How many layers for QAOA (Parameter: p) do you want?"
            },
            "QIRO_reps": {  # number of QIRO reps
                "values": [2, 3, 4, 5],
                "description": "How QIRO reps (Parameter: n) do you want? Choose this parameter sensibly in relation to"
                               " the graph size!"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            shots: int
            depth: int
            iterations: int
            QIRO_reps: int

        """
        shots: int
        depth: int
        iterations: int
        QIRO_reps: int

    def run(self, mapped_problem: any, device_wrapper: any, config: Config, **kwargs: dict) -> tuple[any, float, dict]:
        """
        Run Qrisp QIRO on the local Qrisp simulator.

        :param mapped_problem: Dictionary with the keys 'graph' and 't'
        :param device_wrapper: Instance of device
        :param config: Solver configuration settings
        :param kwargs: no additionally settings needed, may include the measurement kwargs
        :return: Solution, the time it took to compute it and optional additional information
        """

        n = config["QIRO_reps"]
        p = config["depth"]
        shots = config["shots"]
        max_iter = config["iterations"]

        g = mapped_problem['graph']
        start = start_time_measurement()

        qiro_instance = QIROProblem(g,
                                    replacement_routine=create_max_indep_replacement_routine,
                                    cost_operator=create_max_indep_cost_operator_reduced,
                                    mixer=qiro_rx_mixer,
                                    cl_cost_function=create_max_indep_set_cl_cost_function,
                                    init_function=qiro_init_function
                                    )

        qarg = QuantumVariable(g.number_of_nodes())

        # Run actual optimization algorithm
        try:
            res_qiro = qiro_instance.run_qiro(qarg=qarg, depth=p, n_recursions=n, mes_kwargs={"shots": shots},
                                              max_iter=max_iter)
        except ValueError as e:
            logging.error(f"The following ValueError occurred in module QrispQIRO: {e}")
            logging.error("The benchmarking run terminates with exception.")
            raise Exception("Please refer to the logged error message.") from e
        best_bitstring = self._qiro_select_best_state(res_qiro, create_max_indep_set_cl_cost_function(g))

        def _translate_state_to_nodes(state: str, nodes: list) -> list:
            return [key for index, key in enumerate(nodes) if state[index] == '1']

        winner_state = _translate_state_to_nodes(best_bitstring, nodes=g.nodes())

        return winner_state, end_time_measurement(start), {}

    def _qiro_select_best_state(self, res_qiro, cost_func) -> str:
        """
        This function is used for post_processing, i.e. finding the best solution out of the 10 most likely ones.
        Since QIRO is an optimization algorithm, the most_likely solution can be of bad quality, depending on
        the problem cost landscape.

        :param res_qiro: Dictionary containing the QIRO optimization routine results, i.e. the final state.
        :param cost_func: classical cost function of the problem instance
        """
        maxfive = sorted(res_qiro, key=res_qiro.get, reverse=True)[:10]
        max_cost = 0
        best_state = "0" * len(maxfive[0])

        for key in maxfive:
            if cost_func({key: 1}) < max_cost:
                best_state = key
                max_cost = cost_func({key: 1})

        return best_state



#  Copyright 2021 The QUARK Authors. All Rights Reserved.
#
#  Licensed under the Apache License, Version 2.0 (the "License");
#  you may not use this file except in compliance with the License.
#  You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
#  Unless required by applicable law or agreed to in writing, software
#  distributed under the License is distributed on an "AS IS" BASIS,
#  WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
#  See the License for the specific language governing permissions and
#  limitations under the License.

from typing import TypedDict
import logging

from quark.modules.solvers.Solver import Solver
from quark.modules.Core import Core
from quark.utils import start_time_measurement, end_time_measurement


class Annealer(Solver):
    """
    Class for both quantum and simulated annealing.
    """

    def __init__(self):
        """
        Constructor method.
        """
        super().__init__()
        self.submodule_options = ["Simulated Annealer"]

    def get_default_submodule(self, option: str) -> Core:
        """
        Returns the default submodule based on the provided option.

        :param option: The name of the submodule
        :return: Instance of the default submodule
        """
        if option == "Simulated Annealer":
            from quark.modules.devices.SimulatedAnnealingSampler import SimulatedAnnealingSampler  # pylint: disable=C0415
            return SimulatedAnnealingSampler()
        else:
            raise NotImplementedError(f"Device Option {option}  not implemented")

    def get_parameter_options(self) -> dict:
        """
        Returns the configurable settings for this solver.

        :return: Dictionary of parameter options
        .. code-block:: python

            return {
                    "number_of_reads": {
                        "values": [100, 250, 500, 750, 1000],
                        "description": "How many reads do you need?"
                    }
                }

        """
        return {
            "number_of_reads": {
                "values": [100, 250, 500, 750, 1000],
                "description": "How many reads do you need?"
            }
        }

    class Config(TypedDict):
        """
        Attributes of a valid config.

        .. code-block:: python

            number_of_reads: int

        """
        number_of_reads: int

    def run(self, mapped_problem: dict, device_wrapper: any, config: Config, **kwargs: dict) \
            -> tuple[dict, float, dict]:
        """
        Run the annealing solver.

        :param mapped_problem: Dict with the key 'Q' where its value should be the QUBO
        :param device_wrapper: Annealing device
        :param config: Annealing settings
        :param kwargs: Additional keyword arguments
        :return: Solution, the time it took to compute it and optional additional information
        """

        q = mapped_problem['Q']
        additional_solver_information = {}
        device = device_wrapper.get_device()
        start = start_time_measurement()

        if device_wrapper.device_name != "simulated annealer":
            logging.error("Only simulated annealer available at the moment!")
            logging.error("Please select another solver module.")
            logging.error("The benchmarking run terminates with exception.")
            raise Exception("Please refer to the logged error message.")

        response = device.sample_qubo(q, num_reads=config['number_of_reads'])
        time_to_solve = end_time_measurement(start)

        # Take the result with the lowest energy:
        sample = response.lowest().first.sample
        logging.info(f'Annealing finished in {time_to_solve} ms.')

        return sample, time_to_solve, additional_solver_information
